/*

** FILE NAME:	demo.c

**

** PROJECT ID:	94-0040

**

** TITLE:	File containing example DSPLIB functions

**

** AUTHOR:	Ensigma Ltd.

**

** DESCRIPTION:	This file illustrates a few basic library functions.  One

**		of two sinusoids are filtered out both by an FIR filter and

**		a double precision IIR filter.  The energy is shown to be

**		approximately half of the original.  The filtered signal is

**		windowed and an FFT shows the remaining sinusoid

**		to be at 1/8 the sample rate.

**		Library functions used:

**			InitFir()

**			Fir()

**			InitDIir()

**			DIir()

**			MsPower()

**			Variance()

**			GenHamming()

**			VectorMult()

**			InitFft()

**			FftReal()

**			FreeFft()

**

**

** MODIFICATION HISTORY:

**

**              1.00    15-Sep-95       Initial Revision

**

*/

#include <stdio.h>

#include <math.h>

#include <ensigdsp.h>

#include "demoio.h"



#define		VLEN	128

#define		FLEN	12	/* length of FIR filter */

#define		K	2	/* number of IIR biquads */

#define		RS	15	/* Q15 fractional format shift */



/* filter coefficients for FIR and DIIR filters cutoff at Fs/4 */

short firco[FLEN] =

{107, 250, -732, -1779, 4120, 14418, 14418, 4120, -1779, -732, 250, 107};

long diirco[K * 6] =

{31, 1254686956, -496866304, 347415747, 694831502, 347415746, 47, -113001278, -1523568505, 893094203, 1786188388, 893094206};



void

main()

{

    short *sws;

    long *lws;

    long var, pow;

    short mean;

    char message[100];

    short input[VLEN];

    short output[VLEN];

    short window[VLEN];

    short fftinp[VLEN];

    short fftout[VLEN];



/* generate two sinusoids */

    GetData(input, VLEN);



/* check energy */

    if (MsPower(&pow, input, VLEN) != EDSP_OK)

	Error("MsPower Problem\n");

    sprintf(message, "\nPower of signal: %ld", pow);

    OutRes(message);



/* FIR filter out one sinusoid */

    if (InitFir(&sws, VLEN) != EDSP_OK)

	Error("InitFir Problem\n");

    if (Fir(output, input, VLEN, firco, FLEN, RS, sws) != EDSP_OK)

	Error("Fir Problem\n");



/* check that energy is approximately that of one sinusoid */

    if (Variance(&var, &mean, output, VLEN) != EDSP_OK)

	Error("Variance Problem\n");

    sprintf(message, "\nAfter FIR filtering, Variance: %ld, Mean: %ld", var, mean);

    OutRes(message);



/* IIR filter out one sinusoid with double precision */

    if (InitDIir(&lws, K) != EDSP_OK)

	Error("InitDIir Problem\n");

    if (DIir(output, input, VLEN, diirco, K, lws) != EDSP_OK)

	Error("DIir Problem\n");



/* check that energy is approximately that of one sinusoid */

    if (Variance(&var, &mean, output, VLEN) != EDSP_OK)

	Error("Variance Problem\n");

    sprintf(message, "\nAfter DIIR filtering, Variance: %ld, Mean: %ld", var, mean);

    OutRes(message);



/* window the output in preparation for an FFT */



    if (GenHamming(window, VLEN) != EDSP_OK)

	Error("GenHamming problem\n");



    if (VectorMult(fftinp, output, window, VLEN, RS) != EDSP_OK)

	Error("VectorMult problem\n");



/* do FFT */

    if (InitFft(VLEN) != EDSP_OK)

	Error("InitFft problem\n");

    if (FftReal(fftout, fftinp, VLEN, EFFTALLSCALE) != EDSP_OK)

	Error("FftReal problem\n");



    OutSpect(fftout, VLEN);

    FreeFft();

}

