;
; FILE NAME:    fftc.asm
;
; PROJECT ID:   94-0040
;
; TITLE:        FFT
;
; AUTHOR:       Ensigma Ltd.
;
; DESCRIPTION:  This file contains the main not-in-place FFT. This core
; 		performs a forward, complex transform. The core is also used by
; 		all the other fft rountines to provide the main transform,
; 		with extra front- or back-end processing to perform
;		additional manipulations.
;		A decimation-in-time, radix-2 transform is used. The first 2
;		stages are performed as a single radix-4 stage. The remaining
;		stages use a general butterfly and a special unity twiddle
;		factor stage.
;		Scaling of unity or 1/2 can be specified for each stage.
;

; _FftComplex is the entry point for a complex, forward, not-in-place transform.
;
; Four parameters are required They are:
; 1.output data address
; 2.input data address,
; 3.size of data
; 4.scaling factor.
; These parameters are passed in registers R4, R5, R6, R7.
; The return value of the function is
;	EDSP_OK      if the transform has been performed successfully,
;	EDSP_BAD_ARG if size<4, size>max_fft_size, or size not a power of 2.
;

	.EXPORT	_FftComplex
	.EXPORT	_IfftComplex
	.EXPORT	FFTCBV		; entry point for other FFT routines
	.EXPORT	FFTICBV		; entry point for other FFT routines
	.IMPORT	_twtble
	.IMPORT	_bvtble
	.IMPORT	_max_fft_size

forward	.EQU	0
inverse	.EQU	1

; The following variables are stored on the stack:
; INVERSE	flag set to 0 for a forward transform, 1 for an inverse
; NO		no. complex points to transform
; Wpt
; INADDR	address of the input data
; Woff
; GROUP COUNT
; SCALE
; OUTADDR4
; OUTADDR	address of the output data
; NO		no. complex points to transform
; INADDR	address of the input data
;
; Also, the following registers are stored on the stack:
; PR
; R8
; R9
; R10
; R11
; R12
; R13
; R14

	.SECTION	P,CODE,ALIGN=4

_FftComplex:
	MOV.L	R14,@-R15	;PUSH REGISTERS ONTO STACK
	MOV.L	R13,@-R15
	MOV.L	R12,@-R15
	MOV.L	R11,@-R15
	MOV.L	R10,@-R15
	MOV.L	R9,@-R15
	MOV.L	R8,@-R15
	STS.L	MACH,@-R15
	STS.L	MACL,@-R15
	STS.L	PR,@-R15
	ADD	#-48,R15	;ALLOCATE PARAMETER AREA
	MOV	#forward,R1
	BRA	LB70
	MOV.L	R1,@R15

_IfftComplex:
	MOV.L	R14,@-R15	;PUSH REGISTERS ON STACK
	MOV.L	R13,@-R15
	MOV.L	R12,@-R15
	MOV.L	R11,@-R15
	MOV.L	R10,@-R15
	MOV.L	R9,@-R15
	MOV.L	R8,@-R15
	STS.L	MACH,@-R15
	STS.L	MACL,@-R15
	STS.L	PR,@-R15
	ADD	#-48,R15	;ALLOCATE PARAMETER AREA
	MOV	#inverse,R1
	MOV.L	R1,@R15


LB70:	MOV.L	R5,@(4,R15)	;SAVE INADDR
	MOV.L	R4,@(12,R15)	;SAVE OUTADDR
	MOV.L	R6,@(8,R15)	;SAVE NO

	MOV.L	BVADDR,R1
	JSR	@R1		;CALL the FFT core
	MOV.L	R7,@(20,R15)	;SAVE SCALE

EN:	ADD	#48,R15		;DEALLOCATE PAPAMETER AREA
	LDS.L	@R15+,PR	;POP REGISTERS FROM STACK
	LDS.L	@R15+,MACL
	LDS.L	@R15+,MACH
	MOV.L	@R15+,R8
	MOV.L	@R15+,R9
	MOV.L	@R15+,R10
	MOV.L	@R15+,R11
	MOV.L	@R15+,R12
	MOV.L	@R15+,R13
	RTS			;RETURN TO CALLING C PROGRAM
	MOV.L	@R15+,R14

	.ALIGN	4
BVADDR	.data.l	FFTCBV


; R4=POINTER TO CURRENT INPUT DATA
; R0=TABLE BASE ADDRESS
; R1=INDEX FOR TABLE
; R6=ADDRESS OF LAST NON-ZERO ENTRY IN THE TABLE

; Not-in-place bit-reversal code.
; This code reorders the data so that it is in bit-reversed order. A lookup
; table is used to provide the normal to bit-reversed mapping. This table
; is generated by InitFft(), which must be called before the first FFT is
; performed.

; The bit reversal table may allow larger FFT's than the current one, in which
; case we need to calculate the step size through the table.

; Either this code or the code at FFTICBV is executed, depending on whether
; an in-place or not-in-place transform is required.

	.ALIGN	4
FFTCBV:	MOV.L	@(20,R15),R7	;READ SCALE
	DIV0U
	MOV.L	@(8,R15),R6	;READ NO
	CMP/PL	R6		;R6>0? must check now to avoid division by zero
	MOV.L	@(12,R15),R0	;READ OUTADDR
	BF	ERROR
	MOV.L	MAXN1,R3
	MOV	R6,R2
	MOV.L	@R3,R3
	SHLL16	R2
	MOV.L	@(4,R15),R4	;READ INADDR
	DIV1	R2,R3		;PERFORM (MAX/NO OF INPUT DATA)
	MOV.L	R6,@(40,R15)	;SAVE NO
	DIV1	R2,R3		;PERFORM (MAX/NO OF INPUT DATA)
	MOV.L	TBLE,R7		;R7=TABLE BASE ADDRESS

	DIV1	R2,R3		;PERFORM (MAX/NO OF INPUT DATA)
	MOV.L	@R7,R7

	.AREPEAT	13
	DIV1	R2,R3		;PERFORM (MAX/NO OF INPUT DATA)
	.AENDR
	ROTCL	R3
	EXTU.W	R3,R1		;R1 = INCREMENT IN BIT-REVERSAL TABLE


; Now perform the main argument checking, ie is N a power of 2, too large or
; too small?
	MOV	R6,R2		;R2=NO
	ADD	#-4,R2
	CMP/PZ	R2
	BF	ERROR		;NO < 4
	MOV.L	MAXN1,R3
	CMP/PL	R1
	BF	ERROR2		;NO > MAX
	MULS	R1,R6
	MOV.L	@R3,R3
	STS	MACL,R2
	CMP/EQ	R2,R3
	BF	ERROR1		;NO IS NOT A POWER OF 2

; Do the bit-reversal

	MOV.L	TBLE,R7		;R7=TABLE BASE ADDRESS
	SHLL	R1		;offsets are words

	MOV.L	@R7,R7

	SHLL2	R6		;NO*4
	MOV	#0,R12		;LOOP COUNT
	NOP
bitrev:	MOV.W	@R7,R2		;READ TABLE ENTRY
	SHLL2	R2		;*4
	ADD	R4,R2		;ADD IADDR

	ADD	R1,R7		;INCREASE POINTER
	MOV.L	@R2,R3		;READ FROM INPUT DATA
	MOV.L	R3,@(R0,R12)	;STORE IN OUTPUT AREA
	ADD	#4,R12		;INCREASE LOOP COUNT
	CMP/EQ	R12,R6		;LOOP COUNT = NO ?
	BF	bitrev		;IF NOT, REPEAT PROCESS

	BRA	RI4		;BRANCH TO THE FIRST RADIX-4 FFT STAGE
	MOV.L	@(20,R15),R0	;READ SCALE

; In-place bit-reversal code.
; This code reorders the data so that it is in bit-reversed order. A lookup
; table is used to provide the normal to bit-reversed mapping. This table
; is generated by InitFft(), which must be called before the first FFT is
; performed.

; The bit reversal table may allow larger FFT's than the current one, in which
; case we need to calculate the step size through the table.
; Only one of this code or the code at FFTCBV is executed.

	.ALIGN	4
FFTICBV:
	MOV.L	@(20,R15),R7	;READ SCALE
	DIV0U
	MOV.L	@(8,R15),R6	;READ NO
	CMP/PL	R6		;R6 > 0 ?
	MOV.L	@(12,R15),R5	;READ OUTADDR
	BF	ERROR
	MOV.L	MAXN1,R3
	MOV	R6,R2
	MOV.L	@R3,R3
	SHLL16	R2
	MOV.L	@(4,R15),R4	;READ INADDR
	DIV1	R2,R3		;PERFORM (MAX/NO OF INPUT DATA)
	MOV.L	R6,@(40,R15)	;SAVE NO
	DIV1	R2,R3		;PERFORM (MAX/NO OF INPUT DATA)
	MOV.L	TBLE,R7		;R7=TABLE BASE ADDRESS

	DIV1	R2,R3		;PERFORM (MAX/NO OF INPUT DATA)
	MOV.L	@R7,R7

	.AREPEAT	13
	DIV1	R2,R3		;PERFORM (MAX/NO OF INPUT DATA)
	.AENDR
	ROTCL	R3
	EXTU.W	R3,R1		;RESULT IN R1


; Now perform the main argument checking, ie is N a power of 2, too large or
; too small?
	MOV	R6,R2		;R2=NO
	ADD	#-4,R2
	CMP/PZ	R2
	BF	ERROR		;NO < 4
	MOV.L	MAXN1,R3
	CMP/PL	R1
	MOV.L	@R3,R3
	MULS	R1,R6
	MOV	LASTADDR,R0	;R0=TABLE ADDRESS
	BF	ERROR2		;NO > MAX
	STS	MACL,R2
	CMP/EQ	R2,R3
	BF	ERROR1		;NO IS NOT A POWER OF 2
	SHLL	R1

; Do the bit-reversal

; Find the index of the last element to be swapped, to allow the bit reversal
; loop to be terminated as soon as possible
LB41:	MOV.L	@R0+,R12	;READ TABLE
	CMP/EQ	R12,R6		;NO=TABLE ENTRY ?
	MOV.L	@R0+,R12	;READ DATA
	BF	LB41
	SHLL2	R12		;*4
	MOV	R12,R6
	ADD	R4,R6		;ADDRESS OF LAST DATA TO BE SWAPPED

	MOV	R5,R0		;R0=R5=OUTADDR
	SUB	R4,R0		;R0=OFFSET=OUTADDR-INADDR
	MOV	R4,R12		;R12=R4=IADDR

bitrev1:
	MOV.W	@R7,R2		;READ TABLE ENTRY=ADDRESS OF SWAP DATA
	SHLL2	R2		;*2

	ADD	R12,R2		;ADD IADDR
	CMP/GT	R4,R2		;R2>R4?
	BF	skip		;SKIP IF NOT

	MOV.L	@R2,R3		;SWAP DATA POINTED BY POINTER AND TABLE ENTRY
	MOV.L	@R4,R5
	MOV.L	R3,@(R0,R4)
	MOV.L	R5,@(R0,R2)
skip:	CMP/HS	R6,R4		; R4 > R6 ?
	ADD	R1,R7		;INDEX INTO TABLE
	ADD	#4,R4		;INCREASE POINTER
	BF	bitrev1

	BRA	RI4		;BRANCH TO THE FIRST RADIX-4 FFT STAGE
	MOV.L	@(20,R15),R0	;READ SCALE

ERROR:	RTS
	MOV	#1,R0

ERROR1:	RTS
	MOV	#1,R0

ERROR2:	RTS
	MOV	#1,R0

; List of the last elements in the bit reversal table. This is only used by
; the in-place bitreversal.
	.ALIGN	4
LASTADDR:	.data.l	LASTBITV
LASTBITV:	.data.l	4
	.data.l	1
	.data.l	8	;size
	.data.l	3	;last element
	.data.l	16	;size
	.data.l	11	;last element
	.data.l	32	;etc
	.data.l	23
	.data.l	64
	.data.l	55
	.data.l	128
	.data.l	111
	.data.l	256
	.data.l	239
	.data.l	512
	.data.l	479
	.data.l	1024
	.data.l	991
	.data.l	2048
	.data.l	1983
	.data.l	4096
	.data.l	4031
	.data.l	8192
	.data.l	8063
	.data.l	16384
	.data.l	16255
	.data.l	32768
	.data.l	32511

	.ALIGN	4
MAXN1	.data.l	_max_fft_size

; Register use in the following code:
; R12=N = NO OF INPUT DATA
; R13-> FIRST INPUT DATA OF THE BUTTERFLY
; R10 -> THE LAST OUTPUT DATA OF THE BUTTERFLY

; The data is now in bit-reversed order.
; Next perform the preprocessing required for an inverse transform:
; An inverse FFT is equivalent to a forward FFT whose input and output data are
; replaced by their complex conjugates.
; So if this is an inverse transform, replace the data by its complex conjugage.
	.ALIGN 4
	MOV.L	@(20,R15),R0	;READ SCALE
RI4:	MOV	#1,R9
	MOV.L	@(8,R15),R12	;READ NO OF COMPLEX INPUT DATA
	MOV	R12,R11		;R11=R12=N
	MOV.L	@(12,R15),R10	;READ OUTADDR
	MOV	R10,R13		;R10=R13=OUTPUT ADDRESS
	MOV.L	@R15,R1		;R1=inverse
	TST	R9,R1		;inverse=1 ?
	BT	RX4		;Otherwise skip this conjugation stage

	MOV	R12,R2		;R2=R12=COUNTER
	ADD	#2,R13
	NOP
CONJ:	MOV.W	@R13,R1		;READ IMAGINARY PART OF INPUT DATA
	NEG	R1,R1		;NEGATE IT
	MOV.W	R1,@R13		;STORE BACK
	ADD	#4,R13		;INCREASE POINTER
	ADD	#-1,R12		;DECREASE COUNTER
	CMP/PL	R12
	BT	CONJ

	MOV	R10,R13		;R13 -> INPUT DATA
	MOV	R2,R12		;RESUME COUNTER

; Now at the beginning of a forward, in-place, complex transform -
;	bit reversal has been performed,
;	inverse to forward mapping has been performed,
;	real to complex mapping for inverse transforms has been performed.
;
; The first two radix-2 stages are always performed as a single radix-4 stage
; (this is why the minimum size for a complex transform is 4 points).

; The radix-4 stage performs the processing of two radix-2 stages, so there
; are three possible scalings required -
; no scaling, /2 scaling or /4 scaling.
; The main radix-4 loop is repeated for these three cases, but with different
; scaling.

; RADIX-4 BUTTERFLY
;        I/P                 O/P
;
; R13 -> R1     R1+R3           R1+R3+R5+R7
;        R2     R2+R4           R2+R4+R6+R8
;
;        R3     R1-R3           R1-R3+R6-R8
;        R4     R2-R4           R2-R4-R5+R7
;
;        R5     R5+R7           R1+R3-R5-R7
;        R6     R6+R8           R2+R4-R6-R8
;
;        R7     R5-R7           R1-R3-R6+R8
;        R8     R6-R8           R2-R4+R5-R7   <- R10
;
; For this algorithms the 4 inputs to each radix-4 butterfly are stored in
; consecutive locations.

RX4:	ADD	#14,R10		;R10= OUTPUT ADDRESS+14
	SHLR	R0		;SHIFT SCALE FACTOR
	BF	LB20
	SHLR	R0
	BT	RADIX4B		;SCALE FACTOR=4
	BRA	RADIX4A1	;SCALE FACTOR=2
	MOV.W	@R13+,R1	;READ 4 COMPLEX SAMPLES INTO BUTTERFLY
LB20:	SHLR	R0
	BT	RADIX4A		;SCALE FACTOR=2
	BRA	RADIX41
	MOV.W	@R13+,R1	;READ 4 COMPLEX SAMPLES INTO BUTTERFLY

;
; radix4 butterfly with unity scaling
;
	.ALIGN	4
RADIX4:	MOV.W	@R13+,R1	;READ 4 COMPLEX SAMPLES INTO BUTTERFLY
RADIX41:	MOV	R1,R9	;R9=R1
	MOV.W	@R13+,R2
	MOV	R2,R14		;R14=R2
	MOV.W	@R13+,R3
	ADD	R3,R1		;R1=R1+R3
	MOV.W	@R13+,R4
	ADD	R4,R2		;R2=R2+R4
	MOV.W	@R13+,R5
	SUB	R3,R9		;R9=R1-R3
	MOV.W	@R13+,R6
	SUB	R4,R14		;R14=R2-R4
	MOV.W	@R13+,R7
	MOV	R5,R3
	MOV.W	@R13+,R8

	MOV	R6,R4
	ADD	R7,R5		;R5=R5+R7
	ADD	R8,R6		;R6=R6+R8
	SUB	R7,R3		;R3=R5-R7
	SUB	R8,R4		;R4=R6-R8

	MOV	R14,R7		;R7=R14
	MOV	R9,R8		;R8=R9
	ADD	R3,R14
	SUB	R4,R9
	MOV.W	R14,@R10
	MOV	R1,R14		;R14=R1
	MOV.W	R9,@-R10

	MOV	R2,R9		;R9=R2
	SUB	R6,R2
	SUB	R5,R1
	MOV.W	R2,@-R10

	SUB	R3,R7
	MOV.W	R1,@-R10

	ADD	R4,R8
	MOV.W	R7,@-R10

	ADD	R6,R9
	MOV.W	R8,@-R10

	ADD	R5,R14
	MOV.W	R9,@-R10

	ADD	#-4,R12		;DECREASE COUNTER
	MOV.W	R14,@-R10

	ADD	#30,R10		;ADJUST OUTPUT POINTER
	CMP/PL	R12		;COUNTER>0 ?
	BT	RADIX4		;IF NOT, NEXT ROUND

	BRA	RADIX2		;BRANCH TO RADIX2
	MOV	#1,R9		;R9=1

;
; radix4 with scaling factor 1/2
;
	.ALIGN	4
RADIX4A:	MOV.W	@R13+,R1	;READ 4 COMPLEX SAMPLES INTO BUTTERFLY
RADIX4A1:	MOV	R1,R9		;R9=R1
	MOV.W	@R13+,R2
	MOV	R2,R14		;R14=R2
	MOV.W	@R13+,R3
	ADD	R3,R1		;R1=R1+R3
	MOV.W	@R13+,R4
	ADD	R4,R2		;R2=R2+R4
	MOV.W	@R13+,R5
	SUB	R3,R9		;R9=R1-R3
	MOV.W	@R13+,R6
	SUB	R4,R14		;R14=R2-R4
	MOV.W	@R13+,R7
	MOV	R5,R3
	MOV.W	@R13+,R8

	MOV	R6,R4
	ADD	R7,R5		;R5=R5+R7
	ADD	R8,R6		;R6=R6+R8
	SUB	R7,R3		;R3=R5-R7
	SUB	R8,R4		;R4=R6-R8

	MOV	R14,R7		;R7=R14
	MOV	R9,R8		;R8=R9
	ADD	R3,R14
	SHAR	R14		;/2
	MOV.W	R14,@R10
	SUB	R4,R9
	SHAR	R9		;/2
	MOV	R1,R14		;R14=R1
	MOV.W	R9,@-R10

	MOV	R2,R9		;R9=R2
	SUB	R6,R2
	SHAR	R2		;/2
	MOV.W	R2,@-R10
	SUB	R5,R1
	SHAR	R1		;/2

	SUB	R3,R7
	MOV.W	R1,@-R10

	SHAR	R7		;/2
	MOV.W	R7,@-R10
	ADD	R4,R8
	SHAR	R8		;/2

	ADD	R6,R9
	MOV.W	R8,@-R10
	SHAR	R9		;/2
	MOV.W	R9,@-R10

	ADD	R5,R14
	SHAR	R14		;/2
	ADD	#-4,R12		;DECREASE COUNTER
	MOV.W	R14,@-R10

	ADD	#30,R10		;ADJUST OUTPUT POINTER
	CMP/PL	R12		;COUNTER>0 ?
	BT	RADIX4A		;IF NOT, NEXT ROUND

	BRA	RADIX2
	MOV	#1,R9		;R9=1

;
; radix4 with scaling factor 1/4
;
	.ALIGN	4
RADIX4B:	MOV.W	@R13+,R1	;READ 4 COMPLEX SAMPLES INTO BUTTERFLY
	MOV	R1,R9		;R9=R1
	MOV.W	@R13+,R2
	MOV	R2,R14		;R14=R2
	MOV.W	@R13+,R3
	ADD	R3,R1		;R1=R1+R3
	MOV.W	@R13+,R4
	ADD	R4,R2		;R2=R2+R4
	MOV.W	@R13+,R5
	SUB	R3,R9		;R9=R1-R3
	MOV.W	@R13+,R6
	SUB	R4,R14		;R14=R2-R4
	MOV.W	@R13+,R7
	MOV	R5,R3
	MOV.W	@R13+,R8

	MOV	R6,R4
	ADD	R7,R5		;R5=R5+R7
	ADD	R8,R6		;R6=R6+R8
	SUB	R7,R3		;R3=R5-R7
	SUB	R8,R4		;R4=R6-R8

	MOV	R14,R7		;R7=R14
	MOV	R9,R8		;R8=R9
	ADD	R3,R14
	ADD	#2,R14		;ADD ROUNDING
	SHLR2	R14		;/4
	SUB	R4,R9
	MOV.W	R14,@R10

	ADD	#2,R9		;ADD ROUNDING
	SHLR2	R9		;/4
	MOV	R1,R14		;R14=R1
	MOV.W	R9,@-R10

	MOV	R2,R9		;R9=R2
	SUB	R6,R2
	ADD	#2,R2		;ADD ROUNDING
	SHLR2	R2		;/4
	SUB	R5,R1
	MOV.W	R2,@-R10

	ADD	#2,R1		;ADD ROUNDING
	SHLR2	R1		;/4
	SUB	R3,R7
	MOV.W	R1,@-R10

	ADD	#2,R7		;ADD ROUNDING
	SHLR2	R7		;/4
	ADD	R4,R8
	MOV.W	R7,@-R10

	ADD	#2,R8		;ADD ROUNDING
	SHLR2	R8		;/4
	ADD	R6,R9
	MOV.W	R8,@-R10

	ADD	#2,R9		;ADD ROUNDING
	SHLR2	R9		;/4
	ADD	R5,R14
	MOV.W	R9,@-R10

	ADD	#2,R14		;ADD ROUNDING
	SHLR2	R14		;/4
	ADD	#-4,R12		;DECREASE COUNTER
	MOV.W	R14,@-R10

	ADD	#30,R10		;ADJUST OUTPUT POINTER
	CMP/PL	R12		;COUNTER>0 ?
	BT	RADIX4B		;IF NOT, NEXT ROUND

	BRA	RADIX2
	MOV	#1,R9		;R9=1

; Have now calculated the first 2 radix-2 stages (as a single radix-4 stage)
; The remaining stages are performed as groups of radix-2 butterflies.
; For each stage scaling of x1 or x1/2 is allowed.

; RADIX-2 BUTTERFLY
;
;             A      A+B*W
;              \   /       
;                X
;              /   \
;             B      A-B*W
;
;
; A=Ar+jAi
; B=Br+jBi
; W=Wr+jWi
;

; Each stage is divided into a number of buttefly groups. Within each group
; all butterflies have the same twiddle factor.
; In each stage the first group is processed with code optimised for a unity
; twiddle factor. The remaining butteflies use a general 4 multiply butterfly.

; In the butterfly code, the main uses of registers are as follows:
;
; R0=4*SPAN OF THE BUTTERFLY
; R1,R2,R6,R8,R12 are used for butterfly calculation.
; R4=Wr			;REAL PART OF TWIDDLE FACTOR
; R5=Wi			;IMAGINARY PART OF TWIDDLE FACTOR
; R7=		        ;INPUT DATA OFFSET
; R9=4
; R11=0X4000		;FOR ROUNDING
; R13 -> Ar		;pointer to Ar
; R14=0X8000000+4*GROUP COUNT-1	; no. butteflies left in group
; R3 no. butterflies in group
; R10 holds total no. butterflies per group in current stage

; Perform the initialisation required for each stage.
	.ALIGN	4
	MOV	#1,R9		;R9=1
RADIX2:	MOV.L	MAXN2,R1
	MOV.L	@R1,R1
	SHLR	R1
	MOV.L	WADDR,R2	;READ R2=ADDRESS OF Wpt
	ADD	#-4,R1		;WOFF=4*(MAX/8-1)
	MOV.L	@R2,R2		;READ R2=Wpt

	ADD	R1,R2		;INCREASE Wpt
	MOV.L	@(8,R15),R3	;READ NO
	ADD	#4,R2

	MOV.W	@R2+,R4		;READ Wr
	SHLR2	R3
	MOV.W	@R2+,R5		;READ Wi
	SHLR	R3		;R3=GROUP COUNT=N/8

	MOV.L	LEAST,R14	;R14=MOST NEGATIVE NO

	ADD	R3,R14		;+GROUP COUNT-1
	SUBV	R9,R14		;R14=GROUP COUNTER
	BT	RET1		;IF UNDERFLOW, FINISH
	MOV.W	ROUND,R11	;R11=ROUNDING
	MOV	R14,R10		;R10=R14
	MOV.L	@(12,R15),R13	;READ INPUT DATA ADDRESS
	SHLR	R0		;IS SCALE FACTOR SET ?
	MOV.L	R3,@(8,R15)	;SAVE NO
	ADD	#4,R13		;R13=INPUT DATA ADDRESS+4

	MOV.L	R13,@(16,R15)	;SAVE INPUT DATA ADDRESS+4

; There are 2 different butterfly outer loops, depending on the scaling
; required. For unity scaling, UNITBUTFY and BUTTERFLY are required. For
; x1/2 scaling, UNITYBUTFY and BUTTERFLYA are required.

; The looping for these butteflies is controlled by two registers, the
; group element count (R9) and the group count (R14). 
; Group element count counts the number of input element in a group of
; butterflies.
; The group count is counting the number of group in a stage.
; The W (twiddle) values are stored in a look-up table and retrieved using
; pointer Wpt (R3) and index pointer Woff (R7).

; Choose the divide-by-two butterfly if required.

	MOV.L	R0,@(20,R15)	;SAVE SCALE
	MOV.L	R3,@(24,R15)	;SAVE GROUP COUNT
	MOV.W	SPAN,R0		;R0=4*SPAN
	MOV.L	R1,@(28,R15)	;SAVE Woff
	MOV	R0,R7
	MOV.L	R13,@(32,R15)	;SAVE INADDR
	ADD	#-4,R13		;R13 -> INPUT DATA
	MOV.L	R2,@(36,R15)	;SAVE Wpt
	MOV	R0,R3		;RESET GROUP ELEMENT COUNT
	ADD	#-4,R3		;DECREASE GROUP ELEMENT COUNT

	BT	R2HALVE		;BRANCH IF SCALE FACTOR SET
	SHLL	R7		;R7=2*R0
	NOP

;
; radix-2 butterfly with unity scaling factor
;

; unity twiddle factor butterflies - this is the first group in each stage
UNITBUTFY	MOV.W	@(R0,R13),R1	;R1=Br
UNITBUTFY1:	MOV.W	@R13+,R12	;R12=Ar
	MOV.W	@R13,R8		;R8=Ai
	MOV	R8,R6
	MOV.W	@(R0,R13),R2	;R2=Bi
	SUB	R2,R8		;Ai-Bi
	MOV.W	R8,@(R0,R13)
	ADD	R2,R6		;Ai+Bi
	MOV.W	R6,@R13
	MOV	R12,R6
	ADD	R1,R12		;Ar+Br
	SUB	R1,R6		;Ar-Br
	MOV.W	R12,@-R13
	SUBV	R9,R14		;DECREASE GROUP COUNTER
	MOV.W	R6,@(R0,R13)
	ADD	R7,R13		;INCREASE INPUT ADDRESS POINTER
	BF	UNITBUTFY	;IF NOT UNDERFLOW, NEXT ROUND
	MOV	R10,R14		;RESET GROUP COUNTER
	MOV.L	@(16,R15),R13	;R13 -> NEXT INPUT DATA
	NOP

;
; general radix2 butterfly with unity scaling factor - ie the remaining groups
;

; COMPLEX MULTIPLICATION OF B*W
BUTTERFLY	MOV.W	@(R0,R13),R1	;R1=Br
	MULS	R1,R4		;
	MOV.W	@R13+,R12	;R12=Ar
	STS	MACL,R6		;Br*Wr
	MOV.W	@(R0,R13),R2	;R2=Bi

	MULS	R2,R5		;Bi*Wi
	STS	MACL,R8
	MULS	R2,R4
	ADD	R8,R6		;R6=Br*Wr+Bi*Wi
	STS	MACL,R2		;Bi*Wr
	MULS	R1,R5		;Br*Wi
	ADD	R11,R6		;ADD ROUNDING
	STS	MACL,R1
	SUB	R1,R2		;R2=Bi*Wr-Br*Wi

; (R6+ROUNDING)>>15, (R4+ROUNDING)>>15
	SHAL	R6		;R6<<1
	SHLR16	R6		;R6>>16

	ADD	R11,R2		;ADD ROUNDING
	SHAL	R2		;R2<<1

; R2/R6 CONTAIN BW
; NOW ADD A AND STORE RESULT
	MOV.W	@R13,R8		;R8=Ai
	SHLR16	R2		;R2>>16
	MOV	R8,R1		;R1=R8=Ai

	SUB	R2,R1		;R1=Ai-(Bi*Wr-Br*Wi)
	MOV.W	R1,@(R0,R13)	;STORE RESULT

	ADD	R2,R8		;R8=Ai+(Bi*Wr-Br*Wi)
	MOV.W	R8,@R13		;STORE RESULT

	MOV	R12,R1		;R1=R12=Ar

	ADD	R6,R12		;R6=Ar+(Br*Wr+Bi*Wi)
	SUB	R6,R1		;R1=Ar-(Br*Wr+Bi*Wi)
	MOV.W	R12,@-R13	;STORE RESULT
	SUBV	R9,R14		;DECREASE GROUP COUNT
	MOV.W	R1,@(R0,R13)	;STORE RESULT

	ADD	R7,R13		;INCREASE INPUT ADDRESS POINTER
	BF	BUTTERFLY

; finished group

; R0=4*SPAN ,SPAN=WIDTH OF BUTTERFLY-1
; SPAN=2**(K-1)	, K=1,2,..=THE NO OF STAGE
; R3=4*(2**(K-1))=4*GROUP ELEMENT COUNT
; R14=4*(2**N/2**K)=4*GROUP COUNT
; W VALUES ARE IN A LOOK-UP TABLE
; R10=GROUP COUNT VALUE=0X8000000-GROUP COUNT=> R14

; reset pointers ready for the next group

	MOV	R10,R14		;RESET GROUP COUNT
	MOV.L	@(32,R15),R13	;READ INADDR,BUTTERFLY POINTER
	ADD	#4,R13		;INCREASE POINTER
	MOV.L	@(36,R15),R2	;READ Wpt
	ADD	#-4,R3		;DECREASE GROUP ELEMENT COUNT
	MOV.L	@(28,R15),R1	;READ Woff

	ADD	R1,R2		;R2=Wpt+Woff=ADDRESS OF NEXT W
	MOV.W	@R2+,R4		;READ NEXT Wr
	CMP/PL	R3		;			GROUP ELEMENT COUNT>0 ?
	MOV.W	@R2+,R5		;READ Wi

	MOV.L	R13,@(32,R15)	;SAVE INADDR
	MOV.L	R2,@(36,R15)	;SAVE Wpt

	BT	BUTTERFLY	;	      	    IF SO, START THE NEXT GROUP

	BRA	NS1		;JUMP TO NEXT STAGE
	MOV.L	@(24,R15),R6	;READ GROUP COUNT

RET1	BRA	FINAL
	MOV.L	@R15,R1

	.ALIGN	4
LEAST	.data.l H'80000000
TBLE	.data.l _bvtble			;ADDRESS OF BIT REVERSAL TABLE
WADDR	.data.l	_twtble			;W TABLE ADDRESS
MAXN2	.data.l	_max_fft_size
SPAN	.data.w	H'0010		;4*4
ROUND	.data.w H'4000		;FOR ROUNDING


BUTTERFYB	BRA	UNITBUTFY1
	MOV.W	@(R0,R13),R1	;R1=Br


;
; radix2 butterfly with scaling factor 1/2
;
	.ALIGN	4
R2HALVE:
	SHLL	R7		;R7=2*R0
	NOP

; unity twiddle factor butterflies - this is the first group in each stage
UNITBUTFYA	MOV.W	@(R0,R13),R1	;R1=Br
	MOV.W	@R13+,R12	;R12=Ar
	MOV.W	@R13,R8		;R8=Ai
	MOV	R8,R6
	MOV.W	@(R0,R13),R2	;R2=Bi
	SUB	R2,R8		;Ai-Bi
	SHAR	R8		;/2
	ADD	R2,R6		;Ai+Bi
	MOV.W	R8,@(R0,R13)
	SHAR	R6		;/2
	MOV.W	R6,@R13
	MOV	R12,R6
	ADD	R1,R12		;Ar+Br
	SHAR	R12		;/2
	MOV.W	R12,@-R13
	SUB	R1,R6		;Ar-Br
	SHAR	R6		;/2
	SUBV	R9,R14		;DECREASE GROUP COUNT
	MOV.W	R6,@(R0,R13)

	ADD	R7,R13		;INCREASE INPUT ADDRESS POINTER
	BF	UNITBUTFYA	;REPEAT

	MOV	R10,R14		;RESET GROUP COUNTER
	MOV.L	@(16,R15),R13	;R13 -> NEXT INPUT DATA
	NOP

;
; general radix2 butterfly with scaling factor 1/2 - ie the remaining groups
;

; The the looping for these butteflies is controlled by two registers, the
; group element count (R9) and the group count (R14).
; Group element count counts the number of input element in a group of
; butterflies.
; The group count is counting the number of group in a stage.
; The W (twiddle) values are stored in a look-up table and retrieved using
; pointer Wpt (R3) and index pointer Woff (R7).

; COMPLEX MULTIPLICATION OF B*W
BUTTERFLYA	MOV.W	@(R0,R13),R1	;R1=Br
	MULS	R1,R4		;
	MOV.W	@R13+,R12	;R12=Ar
	STS	MACL,R6		;Br*Wr
	MOV.W	@(R0,R13),R2	;R2=Bi

	MULS	R2,R5		;Bi*Wi
	STS	MACL,R8
	ADD	R8,R6		;R6=Br*Wr+Bi*Wi

	MOV.W	@R13,R8		;R8=Ai
	MULS	R2,R4
	ADD	#1,R8		;ADD ROUNDING
	STS	MACL,R2		;Bi*Wr
	SHLL16	R8		;Ai<<16
	MULS	R1,R5		;Br*Wi
	SHAR	R8		;Ai>>1
	STS	MACL,R1
	SUB	R1,R2		;R2=Bi*Wr-Br*Wi

; R2/R6 CONTAIN BW
; NOW ADD A AND STORE RESULT
	MOV	R8,R1		;R1=R8=Ai
	SUB	R2,R1		;R1=Ai-(Bi*Wr-Br*Wi)
	SHLR16	R1		;R1 >> 16
	MOV.W	R1,@(R0,R13)	;STORE RESULT

	ADD	R2,R8		;R8=Ai+(Bi*Wr-Br*Wi)
	SHLR16	R8		;>> 16
	ADD	#1,R12		;Ar+ROUNDING
	MOV.W	R8,@R13		;STORE RESULT

	SHLL16	R12		;<< 16
	SHAR	R12		;>> 1
	MOV	R12,R1		;R1=R12=Ar

	ADD	R6,R12		;R6=Ar+(Br*Wr+Bi*Wi)
	SHLR16	R12		;>> 16
	MOV.W	R12,@-R13	;STORE RESULT

	SUB	R6,R1		;R1=Ar-(Br*Wr+Bi*Wi)
	SHLR16	R1		;>> 16
	SUBV	R9,R14		;DECREASE GROUP COUNT
	MOV.W	R1,@(R0,R13)	;STORE RESULT

	ADD	R7,R13		;INCREASE INPUT ADDRESS POINTER
	BF	BUTTERFLYA	;REPEAT


; reset pointers ready for the next group

	MOV	R10,R14		;RESET GROUP COUNT
	MOV.L	@(32,R15),R13	;READ INADDR,BUTTERFLY POINTER
	ADD	#4,R13		;INCREASE POINTER
	MOV.L	@(28,R15),R1	;READ Woff

	MOV.L	@(36,R15),R2	;READ Wpt

	ADD	R1,R2		;R2=Wpt+Woff=ADDRESS OF NEXT W
	MOV.W	@R2+,R4		;READ NEXT Wr
	MOV.W	@R2+,R5		;READ Wi
	ADD	#-4,R3		;DECREASE GROUP ELEMENT 
	MOV.L	R13,@(32,R15)	;SAVE INADDR
	CMP/PL	R3		;			GROUP ELEMENT COUNT>0 ?
	MOV.L	R2,@(36,R15)	;SAVE Wpt

COUNT
	BT	BUTTERFLYA	;	      IF SO,NEXT BUTTERFLY IN THE GROUP

;
; reset pointers ready for the next stage
;
	MOV.L	@(24,R15),R6	;READ GROUP COUNT
NS1:	SHLR	R6		;HALF THE GROUP COUNT
	MOV.L	@(20,R15),R8	;READ SCALE
	SHLR	R1		;HALF THE Woff
	MOV.L	WADDR1,R2	;W TABLE ADDRESS
	ADD	#-2,R1		;(R1+4)/2-4=R1/2-2
	MOV.L	@R2,R2

	ADD	R1,R2		;INCREASE POINTER
	MOV.L	R6,@(24,R15)	;SAVE GROUP COUNT
	ADD	#4,R2
	MOV.L	R1,@(28,R15)	;SAVE Woff
	SHLL	R0		;DOUBLE THE SPAN
	MOV.W	@R2+,R4		;READ Wr
	SHLL	R7		;INPUT DATA OFFSET=2*R0
	MOV.W	@R2+,R5		;READ Wi
	SHLR	R8		;SHIFT SCALE FACTOR
	MOV.L	@(16,R15),R13	;R13=INPUT ADDRESS + 4
	MOV	R0,R3		;RESET GROUP ELEMENT COUNT
	MOV.L	R8,@(20,R15)	;SAVE SCALE
	ADD	#-4,R3		;-1
	MOV.L	R13,@(32,R15)	;SAVE INADDR
	ADD	#-4,R13		;R13=INPUT ADDRESS
	MOV.L	R2,@(36,R15)	;SAVE Wpt

	BT	SHT1		;BRANCH IF SCALE FACTOR IS SET

	MOV.L	LEAST1,R14
	ADD	R6,R14
	SUBV	R9,R14		;SET NEW GROUP COUNT
	MOV	R14,R10
;IF GROUP COUNT != 0, GO TO THE NEXT NON-SCALING STAGE
	BF	BUTTERFYB	

	BRA	FINAL		;BRANCH END
	MOV.L	@R15,R1

	.ALIGN	4
SHT1	MOV.L	LEAST1,R14
	ADD	R6,R14
	SUBV	R9,R14		;SET NEW GROUP COUNT
	MOV	R14,R10
;IF GROUP COUNT != 0, GO TO THE NEXT SCALING STAGE
	BF	UNITBUTFYA


;
; Have now performed all butterfly processing. However, if this is an inverse
; transform the data must be complex conjugated before the routine returns.
;

	MOV.L	@R15,R1
FINAL:	MOV.L	@(12,R15),R13	;R13=OUTADDR
	TST	R9,R1		;inverse?
	MOV.L	@(40,R15),R12	;READ INPUT DATA NO
	BT	RET		;Skip conjugation if inverse

; IF INVERSE FFT,
; PERFORM COMPLEX CONJUGAGE OF THE OUTPUT DATA
	ADD	#2,R13		;R13-> IMAGINARY PART OF FIRST DATA
	NOP
CONJ2:	MOV.W	@R13,R1		;READ IMAGINARY PART OF INPUT DATA
	NEG	R1,R1		;NEGATE IT
	MOV.W	R1,@R13		;STORE BACK
	ADD	#4,R13		;INCREASE POINTER
	ADD	#-1,R12		;DECREASE COUNTER
	CMP/PL	R12
	BT	CONJ2

RET:	RTS			;RETURN
	MOV	#0,R0


	.ALIGN	4
LEAST1	.data.l	H'80000000
WADDR1	.data.l	_twtble			;W TABLE FOR FFT
SPAN1	.data.w	H'0010

	.END
