/*
** FILE NAME:	fftcore.c
**
** PROJECT ID:	94-0040
**
** TITLE:	FFT in place complex core function, FftCore
**
** AUTHOR:	Ensigma Ltd.
**
** DESCRIPTION:	This file contains the forward, in-place, complex
**		transform that is the core of all the FFT routines.
**		It is derived from many influences, probably originating
**		with the Numerical Recipies code.
**
**		int FftCore(data, size, scale)
**
** PARAMETERS:
**              short data[]        : Complex  data
**              long size           : size of FFT
**              long scale          : scaling specification
**                                          No scaling = 0
**                                          Complete scaling = 0xffffffff
**
** RETURNS:     EDSP_OK       : success
**              EDSP_BAD_ARG  : size < 2
**                              size not a power of 2.
**                              size > max_fft_size
**				
** MODIFICATION HISTORY:
**              1.0  20-7-95    Initial revision
*/

#include <ensigdsp.h>
#define SWAP(a,b) {short tempr=(a);(a)=(b);(b)=tempr;}

int
FftCore(short data[], long size, long scale)
{
    long n, xx, yy, mmax, m, istep, i, j, k;
    int index_scale;
    int factor;
    long round;
    long tempr, tempi;

    if ((size < 4) || (size > max_fft_size))
	return EDSP_BAD_ARG;

    /* Table driven bit reversal */
    index_scale = max_fft_size / size;
    for (i = 0; i < size; i++)
    {
	j = bvtble[i * index_scale];
	if (j < i)
	{
	    SWAP(data[2 * i], data[2 * j]);
	    SWAP(data[2 * i + 1], data[2 * j + 1]);
	}
    }

    n = size << 1;
    mmax = 8;
    yy = (max_fft_size / 8);

    /* Calculate scaling factor for first radix 4 stage */
    if ((scale & 3) == 3)
    {
	factor = 2;
	round = 2;
    }
    else if ((scale & 1) || ((scale >> 1) & 1))
    {
	factor = 1;
	round = 0;
    }
    else
    {
	factor = 0;
	round = 0;
    }

    /*
     * First stage is Radix 4
     */
    for (k = 0; k < 2 * size; k += 8)
    {
	long tempa, tempb, tempc, tempd, tempe, tempf, tempg, temph;

	tempa = (long)data[k] + data[k + 2];
	tempb = (long)data[k] - data[k + 2];

	tempc = (long)data[k + 4] + data[k + 6];
	tempd = (long)data[k + 4] - data[k + 6];

	tempe = (long)data[k + 1] + data[k + 3];
	tempf = (long)data[k + 1] - data[k + 3];

	tempg = (long)data[k + 5] + data[k + 7];
	temph = (long)data[k + 5] - data[k + 7];

	data[k] = (tempa + tempc + round) >> factor;
	data[k + 1] = (tempe + tempg + round) >> factor;
	data[k + 2] = (tempb + temph + round) >> factor;
	data[k + 3] = (tempf - tempd + round) >> factor;
	data[k + 4] = (tempa - tempc + round) >> factor;
	data[k + 5] = (tempe - tempg + round) >> factor;
	data[k + 6] = (tempb - temph + round) >> factor;
	data[k + 7] = (tempf + tempd + round) >> factor;

    }
    scale >>= 2;

    /*
     * remaining stages
     */

    while (n > mmax)
    {

	round = (scale & 1) ? 0x8000L : 0x4000;
	factor = (scale & 1) ? 16 : 15;
	scale = scale >> 1;

	istep = 2 * mmax;
	xx = 0;
	for (m = 1; m < mmax; m += 2)
	{
	    for (i = m; i <= n; i += istep)
	    {
		j = i + mmax;

		tempr = ((long)twtble[2 * xx] * data[j - 1] + (long)twtble[2 * xx + 1] * data[j]);
		tempi = ((long)twtble[2 * xx] * data[j] - (long)twtble[2 * xx + 1] * data[j - 1]);
		data[j - 1] = (((long)data[i - 1] << 15) - tempr + round) >> factor;
		data[j] = (((long)data[i] << 15) - tempi + round) >> factor;
		data[i - 1] = (((long)data[i - 1] << 15) + tempr + round) >> factor;
		data[i] = (((long)data[i] << 15) + tempi + round) >> factor;
	    }
	    xx += yy;

	}
	yy = yy >> 1;
	mmax = istep;

    }
    return EDSP_OK;
}

#undef SWAP
