;
; FILE NAME:    fftr.asm
;
; PROJECT ID:   94-0040
;
; TITLE:        FftReal, IfftReal
;
; AUTHOR:       Ensigma Ltd.
;
; DESCRIPTION:  This function performs a forward, not-in-place, real FFT.
;		This file contains the warping of data required to allow
;		real data to be transformed using a forward complex in-place
;		transform of half the size. (The warping used here is defined
;		in Numerical Recipes in C.) The common FFT core is provided
;		elsewhere.
;		This code is similar to fftinr.asm - however, the code here
;		contains minor differences to allow an in-place transform to
;		be performed efficiently

; _FftReal is the entry point.
;
; Four parameters are required. They are:
; 1.output data address
; 2.input data address
; 2.size of data
; 3.scaling factor.
; These parameters are passed in registers R4, R5, R6.
; The return value of the function is
;       EDSP_OK      if the transform has been performed successfully,
;       EDSP_BAD_ARG if size<4, size>max_fft_size, or size not a power of 2.
;

	.EXPORT	_FftReal
	.EXPORT	_IfftReal
	.EXPORT	RFFT
	.IMPORT	FFTCBV
	.IMPORT	FFTICBV
	.IMPORT	_twtble
	.IMPORT	_max_fft_size

; Forward/inverse is signalled to the core by a 0 or 1 at the bottom of the
; stack
forward	.EQU	0
inverse	.EQU	1

;PROGRAM ENTRY POINT
;
;THE ORGANISATION OF STACK AREA
; BK
; NO
; Wpt
; INADDR
; Woff
; GROUP COUNT
; SCALE
; OUTADDR4
; OUTADDR
; NO
; INADDR
; PR
; PR
; R8
; R9
; R10
; R11
; R12
; R13
; R14

	.SECTION	P,CODE,ALIGN=4

_FftReal:
	MOV.L	R14,@-R15	;PUSH REGISTERS ON STACK
	MOV.L	R13,@-R15
	MOV.L	R12,@-R15
	MOV.L	R11,@-R15
	MOV.L	R10,@-R15
	MOV.L	R9,@-R15
	MOV.L	R8,@-R15
	STS.L	MACH,@-R15
	STS.L	MACL,@-R15
	STS.L	PR,@-R15
	ADD	#-48,R15	;ALLOCATE PARAMETER AREA
	MOV	#forward,R1
	BRA	LB70
	MOV.L	R1,@R15

_IfftReal:
	MOV.L	R14,@-R15	;PUSH REGISTERS ON STACK
	MOV.L	R13,@-R15
	MOV.L	R12,@-R15
	MOV.L	R11,@-R15
	MOV.L	R10,@-R15
	MOV.L	R9,@-R15
	MOV.L	R8,@-R15
	STS.L	MACH,@-R15
	STS.L	MACL,@-R15
	STS.L	PR,@-R15
	ADD	#-48,R15	;ALLOCATE PARAMETER AREA
	MOV	#inverse,R1
	MOV.L	R1,@R15

LB70:	MOV.L	R5,@(4,R15)	;SAVE INADDR
	MOV.L	R4,@(12,R15)	;SAVE OUTADDR
	MOV.L	R6,@(8,R15)	;SAVE NO
	MOV.L	R6,@(40,R15)	;SAVE NO

; call the main processing function, then return and restore the stack
	MOV.L	RFFTADDR,R1
	JSR	@R1		;CALL FFT
	MOV.L	R7,@(20,R15)	;SAVE SCALE

EN:	ADD	#48,R15		;DEALLOCATE PAPAMETER AREA
	LDS.L	@R15+,PR	;POP REGISTERS FROM STACK
	LDS.L	@R15+,MACL
	LDS.L	@R15+,MACH
	MOV.L	@R15+,R8
	MOV.L	@R15+,R9
	MOV.L	@R15+,R10
	MOV.L	@R15+,R11
	MOV.L	@R15+,R12
	MOV.L	@R15+,R13
	RTS			;RETURN TO CALLING C PROGRAM
	MOV.L	@R15+,R14

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    .ALIGN	4

RFFT:	MOV.L	@R15,R2		;R2=BK
	CMP/PL	R2		;FFT/IFFT ?
	MOV.L	@(8,R15),R1	;READ NO
	BT	LB31

; For a forward real transform, perform a complex FFT of half the size,
; followed by a separating stage
	ROTR	R1		;NO/2
	BT	ERR1		;NO NOT EVEN

	MOV.L	R1,@(8,R15)	;SAVE NO/2
	MOV.L	R1,@(40,R15)	;SAVE NO/2
	STS	PR,R1
	MOV.L	R1,@(44,R15)	;STORE PR
	MOV.L	BVADDR,R1
	JSR	@R1		;CALL FFT IN-PLACE
	NOP
	CMP/EQ	#0,R0
	MOV.L	@(44,R15),R1
	LDS	R1,PR		;LOAD PR
	BT	LB30

ERR:	RTS			;FFT RETURN !=0
	NOP
	NOP			; extra NOP to align to longword boundary

LB30:	MOV.L	@(40,R15),R1	;READ NO/2
	SHAL	R1		;*2
	MOV.L	R1,@(8,R15)	;SAVE NO
	MOV.L	R1,@(40,R15)	;SAVE NO

LB31:	MOV.L	MAXN3,R3
	DIV0U
; Check that the size is > 0, to avoid the trying to do a division by zero.
	MOV.L	@(8,R15),R2	;R2=NO
	CMP/PL	R2		;R2 > 0 ?
	MOV.L	@R3,R3
	BT	LB32
ERR1:	RTS
	MOV	#1,R0
	NOP

LB32:	SHLL16	R2
	.AREPEAT	16
	DIV1	R2,R3
	.AENDR
	ROTCL	R3
	EXTU.W	R3,R14

; Main error trapping
	MOV	R1,R2		;R2=NO
	ADD	#-8,R2
	CMP/PZ	R2
	BF	ERROR		;NO < 4
	MOV.L	MAXN3,R3
	CMP/PL	R14
	BF	ERROR2		;NO > MAX, so too big for lookup tables?
	MULS	R14,R1
	MOV.L	@R3,R3
	STS	MACL,R2
	CMP/EQ	R2,R3
	BF	ERROR1		;NO IS NOT A POWER OF 2

; Arguments have now been checked
	MOV.L	WPT3,R3		;R3=WPT
	SHLL2	R14		;R14=MAX/NO*4=WOFF
	MOV.L	@R3,R3
	ADD	#4,R3

	MOV.W	RND1,R11	;R11=ROUNDING
	ADD	#-4,R14

	MOV.L	@R15,R1		;R1=BK
	CMP/PL	R1		;FFT/IFFT ?
; If an inverse transform, go and do the data shuffling and complex FFT.
	BT	IFFT
	NOP

; If a forward transform, the complex FFT has been performed - now unpack the
; result.
	MOV.L	@(12,R15),R8	;R8 -> FIRST INPUT DATA
	MOV	R8,R9
	MOV.L	@(8,R15),R1
	SHLL	R1		;NO*2
	ADD	R1,R9
	ADD	#-4,R9		;R9 -> 2ND LAST INPUT DATA

	MOV.W	@R8+,R1		;R1=H0R
	MOV	R1,R7		;R7=R1
	MOV.W	@R8,R2		;R2=H0I
	SUB	R2,R7		;R7=H0R-H0I
	MOV.W	R7,@R8		;STORE RESULT
	ADD	R2,R1		;R1=H0R+H0I
	MOV.W	R1,@-R8		;STORE RESULT

	ADD	#4,R8		;INCREMENT POINTER

lunpack:
	MOV.W	@R8+,R1		;R1=H1R
	MOV	R1,R7
	MOV.W	@R8+,R2		;R2=H1I
	MOV	R2,R10
	MOV.W	@R9+,R5		;R5=H35
	ADD	R5,R7		;R7=Re1
	MOV.W	@R9+,R6		;R6=H3I

	SUB	R6,R10		;R10=Im1
	SUB	R5,R1		;R1=Re2'

	ADD	R14,R3		;R3=WPT+WOFF
	MOV.W	@R3+,R5		;R5=Wr

	MULS	R1,R5
	ADD	R6,R2		;R2=Im2'
	STS.L	MACL,R12	;R12=Re2'*Wr
	MOV.W	@R3+,R6		;R6=Wi

	MULS	R2,R6
	STS.L	MACL,R13	;R13=Im2'*Wi
	ADD	R13,R12		;R12=Re2'*Wr+Im2'*Wi=Re2
	ADD	R11,R12		;ADD ROUNDING
	MULS	R2,R5
	SHAL	R12		;R12<<1
	STS.L	MACL,R5		;R5=Im2'*Wr
	SHLR16	R12		;R12>>16
	MULS	R1,R6
	EXTS.W	R12,R12
	STS.L	MACL,R13	;R13=Re2'*Wi
	SUB	R13,R5		;R5=Im2'*Wr-Re2'*Wi=Im2
	ADD	R11,R5		;ADD ROUNDING
	SHAL	R5		;R5<<1
	SHLR16	R5		;R5>>16
	EXTS.W	R5,R5
	MOV	R7,R13		;R13=R7=Re1
	NEG	R10,R6		;R6=-R10=-Im1
	ADD	R5,R7		;R7=Re1+Im2
	SUB	R12,R10		;R10=Im1-Re2
	SHAR	R10		;/2
	MOV.W	R10,@-R8	;STORE Im
	SHAR	R7		;/2
	MOV.W	R7,@-R8		;STORE Re
	ADD	#4,R8		;INCREMENT POINTER

	SUB	R12,R6		;R6=-Im1-Re2
	SHAR	R6
	MOV.W	R6,@-R9		;STORE Im
	SUB	R5,R13		;R13=Re1-Im2
	SHAR	R13
	MOV.W	R13,@-R9	;STORE Re
	ADD	#-4,R9		;DECREMENT POINTER

	CMP/EQ	R9,R8		;R8>R9 ?
	BF	lunpack		;NO, REPEAT

	ADD	#2,R8
	MOV.W	@R8,R5
	NEG	R5,R6
	MOV.W	R6,@R8

	BRA	LP9
	MOV	#0,R0

ERROR:	RTS
	MOV	#1,R0

ERROR1:	RTS
	MOV	#1,R0

ERROR2:	RTS
	MOV	#1,R0

; End of forward in-place real FFT code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Start of inverse in-place real FFT code
	.ALIGN  4

IFFT:	MOV.L	@(4,R15),R8	;R8 -> FIRST INPUT DATA
	MOV	R8,R9
	MOV.L	@(8,R15),R1
	SHLL	R1		;NO*2
	ADD	R1,R9
	ADD	#-4,R9		;R9 -> 2ND LAST INPUT DATA
	MOV	@(12,R15),R0	;R0=OUTADDR
	SUB	R8,R0		;R0=OUTADDR-IADDR=OFFSET

	MOV.W	@R8+,R1		;R1=H0R
	MOV	R1,R7		;R7=R1
	MOV.W	@R8,R2		;R2=H0I
	ADD	R2,R1		;R1=H0R+H0I
	SUB	R2,R7		;R7=H0R-H0I
	SHAR	R7		;/2
	MOV.W	R7,@(R0,R8)	;STORE RESULT
	SHAR	R1		;/2
	ADD	#-2,R8
	MOV.W	R1,@(R0,R8)	;STORE RESULT

	ADD	#4,R8		;INCREMENT POINTER
	NOP
; Pack the data ready for a transform of half the size
lpack:	MOV.W	@R8+,R1		;R1=H1R
	MOV	R1,R7
	MOV.W	@R8,R2		;R2=H1I
	MOV	R2,R10
	MOV.W	@R9+,R5		;R5=H3R
	ADD	R5,R7		;R7=Re1
	MOV.W	@R9,R6		;R6=H3I

	SUB	R6,R10		;R10=Im1
	SUB	R5,R1		;R1=Re2'
	ADD	R6,R2		;R2=Im2'

	NEG	R2,R5
	MOV	R1,R2		;R2=H1R-H3R
	MOV	R5,R1		;R1=-(H1I+H3I)
	ADD	R14,R3		;R3=WPT+WOFF
	MOV.W	@R3+,R5		;R5=Wr
	MULS	R1,R5
	MOV.W	@R3+,R6		;R6=Wi

	STS.L	MACL,R12	;R12=Re2'*Wr
	MULS	R2,R6
	STS.L	MACL,R13	;R13=Im2'*Wi
	SUB	R13,R12		;R12=Re2'*Wr-Im2'*Wi=Re2
	ADD	R11,R12		;ADD ROUNDING
	SHAL	R12		;R12<<1
	SHLR16	R12		;R12>>16
	EXTS.W	R12,R12
	MULS	R2,R5
	STS.L	MACL,R5		;R5=Im2'*Wr
	MULS	R1,R6
	STS.L	MACL,R13	;R13=Re2'*Wi

	ADD	R13,R5		;R5=Im2'*Wr+Re2'*Wi=Im2

	ADD	R11,R5		;ADD ROUNDING
	SHAL	R5		;R5<<1
	SHLR16	R5		;R5>>16
	EXTS.W	R5,R5
	MOV	R7,R13		;R13=R7=Re1
	NEG	R10,R6		;R6=-R10=-Im1

	ADD	R12,R7		;R7=Re1+Re2
	ADD	R5,R10		;R10=Im1+Im2

	SHAR	R7		;/2
	SHAR	R10		;/2
	MOV.W	R10,@(R0,R8)	;STORE Im
	ADD	#-2,R8
	MOV.W	R7,@(R0,R8)	;STORE Re
	ADD	#4,R8		;INCREMENT POINTER
	SUB	R12,R13		;R13=Re1-Re2
	ADD	R5,R6		;R6=-Im1+Im2

	SHAR	R13
	SHAR	R6
	MOV.W	R6,@(R0,R9)	;STORE Im
	ADD	#-2,R9
	MOV.W	R13,@(R0,R9)	;STORE Re
	ADD	#-4,R9		;DECREMENT POINTER

	CMP/EQ	R9,R8		;R8>R9 ?
	BF	lpack		;NO, REPEAT

; Packing now complete. Now perform a forward complex tranform of half the size

	MOV.W	@R8,R5
	MOV.W	R5,@(R0,R8)
	ADD	#2,R8
	MOV.W	@R8,R5
	NEG	R5,R6
	MOV.W	R6,@(R0,R8)

LB10:	MOV.L	@(8,R15),R1	;READ NO
	ROTR	R1		;NO/2
	BT	ERROR
	MOV.L	R1,@(8,R15)	;SAVE NO/2
	MOV.L	R1,@(40,R15)	;SAVE NO/2

	MOV.L	@(12,R15),R1	;READ OUTADDR
	MOV.L	R1,@(4,R15)	;SAVE OUTADDR TO IADDR

	STS	PR,R1
	MOV.L	R1,@(44,R15)	;SAVE PR
	MOV.L	BVADDR1,R1
	JSR	@R1		;CALL IN-PLACE FFT
	NOP
	MOV.L	@(44,R15),R1
	LDS	R1,PR		;READ PR

LP9:	RTS
	NOP

	.ALIGN	4
WPT3	.data.l	_twtble
RFFTADDR	.data.l	RFFT
BVADDR	.data.l	FFTCBV
BVADDR1	.data.l	FFTICBV
MAXN3	.data.l	_max_fft_size
RND1	.data.w	H'4000

	.END
