/*
** FILE NAME:   iir.c
**
** PROJECT ID:  94-0040
**
** TITLE:       Iir
**
** AUTHOR:      Ensigma Ltd.
**
** DESCRIPTION: Perform infinite impulse response (IIR) filtering
**
**              int Iir(output, input, no_samples, coeff, no_sections,
**                      workspace)
**
** PARAMETERS:
**              short output[]        : output samples y.
**              const short input[]   : input samples x.
**              long no_samples       : number of samples N to be filtered.
**              const short coeff[]   : filter coefficients.
**                                      _____________________________________
**                                      | a0k  | a1k | a2k | b0k | b1k | b2k|
**                                      -------------------------------------
**              long no_sections      : number of second order filter sections K
**              short *workspace      : variable containing pointer to filter
**                                      memory.
**
** RETURNS:     EDSP_OK               : success
**              EDSP_BAD_ARG          : no_samples  < 1
**                                      no_sections < 1
**                                      a0k         < 0
**                                      a0k         > 16
**
** MODIFICATION HISTORY:
**              1.0  15-9-95    Initial revision
*/

#include <math.h>
#include <ensigdsp.h>

int
Iir(short output[],
    const short input[],
    long no_samples,
    const short coeff[],
    long no_sections,
    short *workspace)
{
    /* variable declaration */
    int j, k, i;
    double temp, accum;

    /* Input Argument Error Checking */
    if ((no_samples < 1) || (no_sections < 1))
	return EDSP_BAD_ARG;

    /* ak0 error checking */
    for (k = 0; k < no_sections; k++)
	if ((coeff[6 * k] > 16) || (coeff[6 * k] < 0))
	    return EDSP_BAD_ARG;

    for (j = 0; j < no_samples; j++)	/* Copy input to output */
	output[j] = input[j];

    for (j = 0; j < no_samples; j++)
    {
	for (k = 0; k < no_sections; k++)
	{
	    /* d(n) = [a2*d(n-2) + a1*d(n-1)  + x(n)<<15]>>15   */
	    temp = ((double)coeff[6 * k + 2] * workspace[k * no_sections + 1] +
		    (double)coeff[6 * k + 1] * workspace[k * no_sections] +
		    (double)output[j] * 32768) / 32768;

	    /* Rounding */
	    temp = floor(temp + 0.5);

	    /* Saturation */
	    if (temp > 32767.0)
		temp = 32767.0;
	    else if (temp < -32768.0)
		temp = -32768.0;

	    /* y(n) = b2*d(n-2) + b1*d(n-1) + b0*d(n)  */
	    accum = (double)coeff[6 * k + 5] * workspace[k * no_sections + 1] +
		(double)coeff[6 * k + 4] * workspace[k * no_sections] +
		(double)coeff[6 * k + 3] * temp;

	    /* d(n-2) = d(n-1)   Delay by one */
	    workspace[k * no_sections + 1] = workspace[k * no_sections];

	    /* d(n-1) = d(n) */
	    workspace[k * no_sections] = (short)temp;

	    /* y(n) = [y(n) + (1 <<(a0 -1))] >> a0  */
	    for (i = 0; i < coeff[6 * k]; i++)
		accum /= 2;

	    accum = floor(accum + 0.5);

	    /* Saturation */
	    if (accum > 32767)
		accum = 32767;
	    else if (accum < -32768)
		accum = -32768;

	    output[j] = accum;
	}
    }

    return EDSP_OK;		/* Return Success */
}
