/*
** FILE NAME:	iir1.c
**
** PROJECT ID:	94-0040
**
** TITLE:	Iir1
**
** AUTHOR:	Ensigma Ltd.
**
** DESCRIPTION:	Perform infinite impulse response (IIR) filtering for
**		one sample
**
**		int Iir1(output, input, coeff, no_sections, workspace)
**
** PARAMETERS:
**              short output[]        : output sample yk-1(n).
**              short input           : input sample x0(n).
**              const short coeff[]   : filter coefficients.
**                                      _____________________________________
**                                      | a0k  | a1k | a2k | b0k | b1k | b2k|
**                                      -------------------------------------
**              long no_sections      : number of second order filter sections K
**              short *workspace      : variable containing pointer to filter
**                                      memory.
**
** RETURNS:     EDSP_OK               : success
**              EDSP_BAD_ARG          : no_sections < 1
**                                      a0k         < 0
**                                      a0k         > 16
**
** MODIFICATION HISTORY:
**              1.0  15-9-95    Initial revision
*/

#include <math.h>
#include <ensigdsp.h>

int
Iir1(short *output,
     short input,
     const short coeff[],
     long no_sections,
     short *workspace)
{
    /* Variable declaration */
    int k, i;
    double accum, temp;

    /* Input Argument Error Checking */
    if (no_sections < 1)
	return EDSP_BAD_ARG;

    /* ak0 error checking */
    for (k = 0; k < no_sections; k++)
	if ((coeff[6 * k] > 16) || (coeff[6 * k] < 0))
	    return EDSP_BAD_ARG;

    *output = input;		/* Copying input to temp. output */

    for (k = 0; k < no_sections; k++)
    {
	/* d(n) = [a2*d(n-2) + a1*d(n-1) + x(n)<<15]>>15   */
	temp = ((double)coeff[6 * k + 2] * workspace[k * no_sections + 1] +
		(double)coeff[6 * k + 1] * workspace[k * no_sections] +
		(double)(*output) * 32768) / 32768;

	temp = floor(temp + 0.5);

	/* Saturation */
	if (temp > 32767.0)
	    temp = 32767.0;
	else if (temp < -32768.0)
	    temp = -32768.0;

	/* y(n) = b2*d(n-2) + b1*d(n-1) + b0*d(n)  */
	accum = (double)coeff[6 * k + 5] * workspace[k * no_sections + 1] +
	    (double)coeff[6 * k + 4] * workspace[k * no_sections] +
	    (double)coeff[6 * k + 3] * temp;

	/* d(n-2) = d(n-1)  */
	workspace[k * no_sections + 1] = workspace[k * no_sections];

	/* d(n-1) = d(n) */
	workspace[k * no_sections] = temp;

	/* y(n) = [y(n) + (1 <<(a0 -1))] >> a0  */
	for (i = 0; i < coeff[6 * k]; i++)
	    accum /= 2;

	accum = floor(accum + 0.5);

	/* Saturation */
	if (accum > 32767.0)
	    accum = 32767.0;
	else if (accum < -32768.0)
	    accum = -32768.0;

	*output = accum;
    }

    return EDSP_OK;		/* Return Success */
}
