/* Commercial in Confidence */
/*
** FILE NAME:	initfft.c
**
** PROJECT ID:	94-0040
**
** TITLE:	FFT lookup table generation routine
**
** AUTHOR:	Ensigma Ltd.
**
** DESCRIPTION:	This file contains the routines used to generate the twiddle
**		and bit-reversal tables used by the C and assembler FFT
**		rountines.
**
** MODIFICATION HISTORY:
**              1.0  20-7-95    Initial revision
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <ensigdsp.h>
#define PI	3.1415926535897932

short *bvtble = NULL;
short *twtble = NULL;
long max_fft_size = 0;

static void GenBitRev(long nn, short z[]);
static void GenTwiddle(long n, short data[]);

/*
** FUNCTION:	InitFft(max_size)
**
** DESCRIPTION:	Allocates lookup tables via malloc() and initialises them
**		with bit-reversal indicees and twiddle factors for the
**		specified FFT size.
**
** RETURNS:	EDSP_OK       : success
**              EDSP_BAD_ARG  : size < 2
**           			size not a power of 2.
**                              size > max_fft_size
**              EDSP_NO_HEAP  : insufficient space available from malloc.
*/
int
InitFft(long max_size)
{
    int test;

    /* Input Argument error testing */
    if ((max_size < 2) || (max_size > 32768))
	return EDSP_BAD_ARG;

    /* test for power of 2 */
    for (test = max_size; test > 2; test /= 2)
	if (test & 1)
	    return EDSP_BAD_ARG;

    /* Bit reversal table */
    if ((bvtble = (short *)malloc(sizeof(short) * max_size)) == NULL)
	 return EDSP_NO_HEAP;

    /* Twiddle factor table */
    if ((twtble = (short *)malloc(sizeof(short) * max_size)) == NULL)
    {
	free(bvtble);
	return EDSP_NO_HEAP;
    }

    GenBitRev(max_size, bvtble);
    GenTwiddle((max_size / 2), twtble);

    /*
     * Initialise the global max. fft size variable
     */
    max_fft_size = max_size;

    return EDSP_OK;		/* return success */
}

/*
** FUNCTION:	FreeFft
**
** DESCRIPTION:	Releases the memory used by FFT lookup tables. After this
**		function has been invoked no more FFT's may be performed unless
**		a further call to InitFft is made.
**
** RETURNS:	void
*/
void
FreeFft(void)
{
    max_fft_size = 0;
    free(bvtble);		/* Free the bvtble array */
    free(twtble);		/* Free the twtble array */
    bvtble = twtble = NULL;
}

/*
** FUNCTION:	GenBitRev
**
** DESCRIPTION:	Calculates the bit-reversal indicees.
**
** RETURNS:	void
*/
static void
GenBitRev(long size, short output[])
{
    short m, bit_rev, normal;

    if (size < 1)
	return;
    output[0] = 0;
    bit_rev = 1;
    /*
     * We assume that size is a power of two
     */
    for (normal = 1; normal < size; normal++)
    {
	m = size / 2;
	while (m >= 1 && bit_rev > m)
	{
	    bit_rev -= m;
	    m >>= 1;
	}
	bit_rev += m;
	output[normal] = bit_rev - 1;
    }
}

/*
** FUNCTION:	GenTwiddle
**
** DESCRIPTION:	Calculates the FFT twiddle factor table.
**
** RETURNS:	void
*/
static void
GenTwiddle(long N, short data[])
{
    long i;
    float real, imag;
    float arg = PI / N;		/* PI/size */
    float wtemp, wr, wpi, wpr, wi;

    /* Variable initial value calculation */
    wtemp = sin(0.5 * arg);
    wpr = -2.0 * wtemp * wtemp;
    wpi = sin(arg);
    wr = 1.0;
    wi = 0.0;

    for (i = 0; i <= N / 2; i++)
    {
	real = 32768.0 * wr + 0.5;	/* cos(i * arg) ; */

	if (real >= 32767.0)
	    data[2 * i] = 32767;
	else
	    data[2 * i] = (short)(real);

	imag = 32768.0 * wi + 0.5;	/* sin(i*arg);        */

	if (imag >= 32767.0)
	    data[2 * i + 1] = 32767;
	else
	    data[2 * i + 1] = (short)(imag);

	/* Trig recurence */
	wr = (wtemp = wr) * wpr - wi * wpi + wr;
	wi = wi * wpr + wtemp * wpi + wi;

    }

    /*
     * Second half of the table is derived from the first half to save
     * computation
     */
    for (i = 1; i <= N / 2; i++)
    {
	data[2 * (N - i)] = (-1) * data[2 * i];
	data[2 * (N - i) + 1] = data[2 * i + 1];

    }
}
