/*
** FILE NAME:   lms.c
**
** PROJECT ID:  94-0040
**
** TITLE:       Lms
**
** AUTHOR:      Ensigma Ltd.
**
** DESCRIPTION: This routine implements a real adaptive FIR filter using the
**              least mean square algorithm.
**
**              int Lms(output, input, ref_output, no_samples, coeff,
**                      no_coeffs, res_shift, conv_fact, workspace)
** PARAMETERS:
**
**               short output[]        : output samples y.
**               const short input[]   : input samples x.
**               const short ref_output: desired output value d.
**               long no_samples       : number of samples N to be filtered.
**               short coeff[]         : adaptive filter coefficients h.
**               long no_coeffs        : number of coefficients K
**               int res_shift         : right shift applied to each output.
**               short conv_fact       : convergence factor 2u
**               short *workspace      : variable containing pointer to filter
**                                       memory.
**
** RETURNS:     EDSP_OK               : success
**              EDSP_BAD_ARG          : no_samples <  1
**                                      no_coeffs  <= 2
**                                      res_shift  <  0
**                                      res_shift  >  27
**
** MODIFICATION HISTORY:
**              1.0  15-9-95    Initial revision
*/

#include <math.h>
#include <ensigdsp.h>

int
Lms(short output[],
    const short input[],
    const short ref_output[],
    long no_samples,
    short coeff[],
    long no_coeffs,
    int res_shift,
    short conv_fact,
    short *workspace)
{
    /* Variable declaration */
    int i, j, next, inc;
    double temp;
    long temp2, newfirst, e;

    /* Input Argument Error Checking */
    if ((no_samples < 1) || (no_coeffs <= 2) || (res_shift < 0) || (res_shift > 27))
	return EDSP_BAD_ARG;

    /* Get the location of the buffer pointer */
    inc = ((((long)workspace[0] << 16) + (long)workspace[1]) - (long)workspace) / 2;

    next = inc;

    /* Copy the input into workspace */
    workspace[inc] = input[0];

    for (j = 0; j < no_samples; j++)	/* Main Loop */
    {
	i = 0;			/* counter initialisation */
	temp = 0;		/* accumulator initialisation */

	while (i < no_coeffs)
	{
	    if (inc - i < 2)	/* check for valid buffer position */
		inc = no_coeffs + i + 1;	/* move to end of buffer */

	    temp += (double)coeff[i] * workspace[inc - i];
	    i++;		/* while counter increment */
	}

	/* Right shifting and rounding */
	temp = floor(temp / (1L << res_shift) + 0.5);

	/* Saturation */
	if (temp > 32767)
	    temp = 32767;
	else if (temp < -32768)
	    temp = -32768;

	output[j] = temp;	/* Extracting the output */

	/* e(n) = d(n) - y(n); */
	e = (long)ref_output[j] - output[j];	/* Compute the error */

	/* Saturation */
	if (e > 32767)
	    e = 32767;
	else if (e < -32768)
	    e = -32768;

	i = 0;			/* While counter initialisation */
	inc = next;		/* buffer pointer initialisation */

	while (i < no_coeffs)
	{
	    if (inc - i < 2)	/* check for valid buffer position */
		inc = no_coeffs + i + 1;	/* move to end of buffer */

	    /* Coefficient update in  two stages */
	    /* Stage 1: temp = [2u *e(n)] >> 15 + 0.5;  */
	    temp = floor((double)conv_fact * e / 32768 + 0.5);

	    /* Stage 2: temp = [temp * x(n-k)] >> 15 + 0.5;  */
	    temp = floor((double)temp * workspace[inc - i] / 32768 + 0.5);

	    /* Coefficient update */
	    temp2 = coeff[i] + temp;
	    coeff[i] = temp2;

	    /* Saturation */
	    if (temp2 > 32767)
		coeff[i] = 32767;
	    else if (temp2 < -32768)
		coeff[i] = -32768;

	    i++;		/* while counter increment */
	}

	/* Pointer to target of next input data copy */
	if (next > no_coeffs)
	    next = 2;
	else
	    next++;

	inc = next;		/* Buffer pointer */

	if (j < (no_samples - 1))
	    workspace[next] = input[j + 1];
    }
    newfirst = (long)&workspace[next];

    workspace[0] = newfirst >> 16;
    workspace[1] = (newfirst & 0xffff);

    return EDSP_OK;		/* Success */
}
