/*
** FILE NAME:   lms1.c
**
** PROJECT ID:  94-0040
**
** TITLE:       Lms1
**
** AUTHOR:      Ensigma Ltd.
**
** DESCRIPTION: This routine implements a real adaptive FIR filter using the
**              least mean square algorithm for a single input sample.
**
**              int Lms1(output, input, ref_output, coeff,
**                      no_coeffs, res_shift, conv_fact, workspace)
** PARAMETERS:
**
**               short output[]        : output sample y.
**               input                 : input sample x.
**               ref_output            : desired output value d.
**               short coeff[]         : adaptive filter coefficients h.
**               long no_coeffs        : number of coefficients K
**               int res_shift         : right shift applied to each output.
**               short conv_fact       : convergence factor 2u
**               short *workspace      : variable containing pointer to filter
**                                       memory.
**
** RETURNS:     EDSP_OK               : success
**              EDSP_BAD_ARG          : no_coeffs  <= 2
**                                      res_shift  <  0
**                                      res_shift  >  27
**
** MODIFICATION HISTORY:
**              1.0  15-9-95    Initial revision
*/

#include <math.h>
#include <ensigdsp.h>

int
Lms1(short *output,
     short input,
     short ref,
     short coeff[],
     long no_coeffs,
     int res_shift,
     short conv_fact,
     short *workspace)
{

    /* Variable declaration */
    int i, inc, inc_store;
    double temp;
    long temp2, newfirst, e;

    /* Error Checking */
    if ((no_coeffs <= 2) || (res_shift < 0) || (res_shift > 27))
	return EDSP_BAD_ARG;

    /* Get the location of the buffer pointer */
    inc = ((((long)workspace[0] << 16) + (long)workspace[1]) - (long)workspace) / 2;
    inc_store = inc;

    /* Copy the input into workspace */

    workspace[inc] = input;

    i = 0;			/* while counter initialisation */
    temp = 0;			/* accum. initialisation */

    while (i < no_coeffs)
    {
	if (inc - i < 2)	/* check for valid buffer position */
	    inc = no_coeffs + i + 1;	/* move to end of buffer */

	temp += (double)coeff[i] * workspace[inc - i];
	i++;			/* while counter increment */
    }

    /* Right shifting and rounding */
    temp = floor(temp / (1L << res_shift) + 0.5);

    /* Saturation */
    if (temp > 32767)
	temp = 32767;
    else if (temp < -32768)
	temp = -32768;

    *output = temp;		/* Extracting the output */

    /* e(n) = d(n) - y(n); */
    e = (long)ref - *output;	/* Compute the error */

    /* Saturation */
    if (e > 32767)
	e = 32767;
    else if (e < -32768)
	e = -32768;

    i = 0;			/* While counter initilisation */
    inc = inc_store;		/* buffer pointer initilisation */

    while (i < no_coeffs)
    {
	if (inc - i < 2)	/* check for valid buffer position */
	    inc = no_coeffs + i + 1;	/* move to end of buffer */

	/* Coefficient update in two stages  */
	/* Stage 1:  temp = [2u *e(n)] >> 15 + 0.5;  */
	temp = floor((double)conv_fact * e / 32768 + 0.5);

	/* Stage 2: temp = [temp * x(n-k)] >> 15 + 0.5;  */
	temp = floor((double)temp * workspace[inc - i] / 32768 + 0.5);

	/* Coefficient update */
	temp2 = coeff[i] + temp;

	coeff[i] = temp2;

	/* Saturation */
	if (temp2 > 32767)
	    coeff[i] = 32767;
	else if (temp2 < -32768)
	    coeff[i] = -32768;

	i++;			/* while counter increment */
    }

    if (inc_store == (no_coeffs + 1))	/* workspace pointer increment */
	newfirst = (long)&workspace[2];
    else
	newfirst = (long)&workspace[inc_store + 1];

    workspace[0] = newfirst >> 16;
    workspace[1] = (newfirst & 0xffff);

    return EDSP_OK;		/* Success */
}
