/*
** FILE NAME: testfft.c
**
** PROJECT ID: 94-0040
**
** TITLE: File containing example test functions
**
** AUTHOR: Ensigma Ltd.
**
** DESCRIPTION: This file contains the tests for the forward FFT functions
**              FftComplex()
**              FftReal()
**              FftInComplex() and
**              FftInReal()
**
** MODIFICATION HISTORY:
**
**              1.00    15-Sep-95       Eddy
**
*/
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "ensigdsp.h"
#include "testhead.h"

#define PI 3.141592654
#define MIN_CFFT_SIZE   4
#define MIN_RFFT_SIZE   8
#define MAX_FFT_SAMP  256       /* size of transform */
#define NO_SCALES       (sizeof(scales)/sizeof(scales[0]))

static int SingleCompare(short *op1, long n, long m, double singval);

static long scales[] =
{EFFTNOSCALE, EFFTMIDSCALE, EFFTALLSCALE};

extern int debug;               /* set to non-zero if debugging */

/*
** FUNCTION:    GetCHalving
**
** DESCRIPTION: Find the overall amplitude scaling that is performed by a
**              complex transform.
*/
static long
GetCHalving(long size, long fft_scale)
{
    int halving;

    for (halving = 1, fft_scale &= (size - 1); fft_scale > 0; fft_scale >>= 1)
	if (fft_scale & 1)
	    halving *= 2;
    return halving;
}

/*
** FUNCTION:    GetRHalving
**
** DESCRIPTION: Find the overall amplitude scaling that is performed by a
**              real transform.
*/
static long
GetRHalving(long size, long fft_scale)
{
    return GetCHalving(size / 2, fft_scale) * 2;
}

/*
** FUNCTION:    TestFft
**
** AUTHOR:      Ensigma Ltd.
**
** DESCRIPTION: This function tests and demonstrates the use of the DSPLib
**              functions:
**                 FftComplex
**                 FftInComplex
**                 FftReal
**                 FftInReal
**              The function generates test signals which produce an output
**              in one FFT bin only. For each signal the test program calls
**              the approriate FFT routine and verifies that all bins except
**              the correct one are approximately zero and that the non-zero
**              bin is of the right size.
**
** PARAMETERS:  rcode       string buffer for status messages
**              n_samp      FFT size to be tested
**              fft_scale   scaling control parameter
**              ip_scale    The amount by which to scale the data for each
**                          transform. This should be chosen to avoid the
**                          possibility of numerical overflow.
**
** RETURNS:     0           success
**              positive    number of failures detected
**              -1          valid arguments were rejected by some function
**
*/

static int
TestOneFft(char * rcode, long n_samp, long fft_scale, long ip_scale)
{
    int i;
    int j;
    double cmplx_scale, real_scale;
    short ip[MAX_FFT_SAMP * 2]; /* input array */
    short op[MAX_FFT_SAMP * 2]; /* normal output array */
    int fail = 0;

#define HALF_CYCLE (n_samp/2)   /* number of samples in half a cycle of the
				 * lowest test frequency */
    if (debug)
	printf("input at %p\noutput at %p\n", ip, op);

    /* Initialise the array to non-zero values */
    for (i = 0; i < MAX_FFT_SAMP * 2; i++)
	ip[i] = op[i] = 0x5555;
    printf("ipscale %lx\nn_samp %ld\nfftscale %lx\n", ip_scale, n_samp, fft_scale);

    /* Find the amplitude scaling that will be performed by a transform */

    cmplx_scale = 1.0 * ip_scale / GetCHalving(n_samp, fft_scale);
    real_scale = 1.0 * ip_scale / GetRHalving(n_samp, fft_scale);

    printf("\nTest complex FFT's\n");
    /*
     * Perform the tests for positive phasors.
     */
    sprintf(rcode, "Testing complex FFT's / positive phasors (size %ld)\n", n_samp);
    for (i = 0; i < HALF_CYCLE; i++)
    {
	/*
	 * positive frequency, 0 phase
	 */
	for (j = 0; j < n_samp; j++)
	{
	    ip[2 * j] = cos(j * i * PI / HALF_CYCLE) * ip_scale;
	    ip[2 * j + 1] = sin(j * i * PI / HALF_CYCLE) * ip_scale;
	}
	if (debug > 255)
	    printf("n_samp=%ld\nCalling FFT\n", n_samp);

	if (FftComplex(op, ip, n_samp, fft_scale) != EDSP_OK)
	    return -1;
	if (debug > 255)
	{
	    printf("output\n");
	    for (j = 0; j < n_samp; j++)
		printf("%hd\t%hd\n", op[2 * j], op[2 * j + 1]);
	    printf("input\n");
	    for (j = 0; j < n_samp; j++)
		printf("%hd\t%hd\n", ip[2 * j], ip[2 * j + 1]);
	}

	if (FftInComplex(ip, n_samp, fft_scale) != EDSP_OK)
	    return -1;
	fail += SingleCompare(op, n_samp * 2, 2 * i, cmplx_scale * n_samp);
	fail += SingleCompare(ip, n_samp * 2, 2 * i, cmplx_scale * n_samp);

	/*
	 * positive frequency, 90 degrees phase
	 */
	for (j = 0; j < n_samp; j++)
	{
	    ip[2 * j] = -sin(j * i * PI / HALF_CYCLE) * ip_scale;
	    ip[2 * j + 1] = cos(j * i * PI / HALF_CYCLE) * ip_scale;
	}
	if (debug > 255)
	{
	    printf("input\n");
	    for (j = 0; j < n_samp; j++)
		printf("%hd\t%hd\n", ip[2 * j], ip[2 * j + 1]);
	}
	if (FftComplex(op, ip, n_samp, fft_scale) != EDSP_OK)
	    return -1;
	if (FftInComplex(ip, n_samp, fft_scale) != EDSP_OK)
	    return -1;
	fail += SingleCompare(op, n_samp * 2, 2 * i + 1, cmplx_scale * n_samp);
	fail += SingleCompare(ip, n_samp * 2, 2 * i + 1, cmplx_scale * n_samp);
    }

    /*
     * Tests for negative phasors.
     */
    sprintf(rcode, "Testing complex FFT's / negative phasors (size %ld)\n", n_samp);
    for (i = 1; i <= HALF_CYCLE; i++)
    {
	/*
	 * negative frequency, 90 degrees phase
	 */
	for (j = 0; j < n_samp; j++)
	{
	    ip[2 * j] = sin(j * i * PI / HALF_CYCLE) * ip_scale;
	    ip[2 * j + 1] = cos(j * i * PI / HALF_CYCLE) * ip_scale;
	}
	if (FftComplex(op, ip, n_samp, fft_scale) != EDSP_OK)
	    return -1;
	if (FftInComplex(ip, n_samp, fft_scale) != EDSP_OK)
	    return -1;
	fail += SingleCompare(op, n_samp * 2, (n_samp - i) * 2 + 1, cmplx_scale * n_samp);
	fail += SingleCompare(ip, n_samp * 2, (n_samp - i) * 2 + 1, cmplx_scale * n_samp);

	/*
	 * negative frequency, 0 phase
	 */
	for (j = 0; j < n_samp; j++)
	{
	    ip[2 * j] = cos(j * i * PI / HALF_CYCLE) * ip_scale;
	    ip[2 * j + 1] = -sin(j * i * PI / HALF_CYCLE) * ip_scale;
	}
	if (FftComplex(op, ip, n_samp, fft_scale) != EDSP_OK)
	    return -1;
	if (FftInComplex(ip, n_samp, fft_scale) != EDSP_OK)
	    return -1;
	fail += SingleCompare(op, n_samp * 2, (n_samp - i) * 2, cmplx_scale * n_samp);
	fail += SingleCompare(ip, n_samp * 2, (n_samp - i) * 2, cmplx_scale * n_samp);
    }

    /*
     * Test the real only transforms.
     */
    if (n_samp >= MIN_RFFT_SIZE)
    {
	/*
	 * Except for the 0 and Fs/2 bins, the signal amplitude must be halved
	 * to prevent overflow
	 */
	if (debug)
	    printf("\nTesting Real transforms\n");
	/*
	 * Real input as 0 phase
	 */
	sprintf(rcode, "Testing real FFT's / 0 phase (size %ld)\n", n_samp);
	for (i = 1; i < HALF_CYCLE; i++)
	{
	    for (j = 0; j < n_samp; j++)
		ip[j] = 1 * cos(j * i * PI / HALF_CYCLE) * ip_scale;
	    if (FftReal(op, ip, n_samp, fft_scale) != EDSP_OK)
		return -1;
	    if (FftInReal(ip, n_samp, fft_scale) != EDSP_OK)
		return -1;
	    fail += SingleCompare(op, n_samp, 2 * i, real_scale * n_samp);
	    fail += SingleCompare(ip, n_samp, 2 * i, real_scale * n_samp);
	}

	/*
	 * Test 0 and Fs/2 frequencies by themselves. First 0 (d.c.):
	 */
	for (j = 0; j < n_samp; j++)
	    ip[j] = ip_scale / 2;
	if (FftReal(op, ip, n_samp, fft_scale) != EDSP_OK)
	    return -1;
	if (FftInReal(ip, n_samp, fft_scale) != EDSP_OK)
	    return -1;
	fail += SingleCompare(op, n_samp, 0, real_scale * n_samp);
	fail += SingleCompare(ip, n_samp, 0, real_scale * n_samp);

	/*
	 * now Fs/2:
	 */
	for (j = 0; j < n_samp; j++)
	    ip[j] = cos(j * HALF_CYCLE * PI / HALF_CYCLE) * ip_scale / 2;
	if (FftReal(op, ip, n_samp, fft_scale) != EDSP_OK)
	    return -1;
	if (FftInReal(ip, n_samp, fft_scale) != EDSP_OK)
	    return -1;
	fail += SingleCompare(op, n_samp, 1, real_scale * n_samp);
	fail += SingleCompare(ip, n_samp, 1, real_scale * n_samp);

	/*
	 * Real input, 90 degrees phase
	 */
	sprintf(rcode, "Testing real FFT's / 90 degrees phase (size %ld)\n", n_samp);
	for (i = 1; i < HALF_CYCLE; i++)
	{
	    for (j = 0; j < n_samp; j++)
		ip[j] = -1 * sin(j * i * PI / HALF_CYCLE) * ip_scale;
	    if (FftReal(op, ip, n_samp, fft_scale) != EDSP_OK)
		return -1;
	    if (FftInReal(ip, n_samp, fft_scale) != EDSP_OK)
		return -1;
	    fail += SingleCompare(op, n_samp, 2 * i + 1, real_scale * n_samp);
	    fail += SingleCompare(ip, n_samp, 2 * i + 1, real_scale * n_samp);
	}
    }

    if (debug)
	printf("TestOneFft returned %d\n", fail);
    return (fail);
}

/*
** FUNCTION:    TestFft
**
** AUTHOR:      Ensigma Ltd.
**
** DESCRIPTION: Test the forward FFT with inputs sized 2^n, from 2 to
**              MAX_FFT_SIZE. Non power of two inputs sizes are also tested
**              (no FFT processing should be performed in this case).
*/
#define SCALE_BASE      0x7ff0

int
TestFft(char *rcode)
{
    long size, j, davescale;
    int scale_idx, result = 0;
    long ipscale;
    short data[MIN_CFFT_SIZE * 2 + MIN_RFFT_SIZE];
    long fftscale;

    /*
     * Check small transforms
     */
    sprintf(rcode, "Testing argument checking of complex FFT's\n");
    for (size = -1; size < MIN_CFFT_SIZE; size++)
    {
	fftscale = 0;
	if (debug > 255)
	    printf("testing complex forward size %ld\n", size);
	if ((FftInComplex(data, size, fftscale) != EDSP_BAD_ARG) ||
	    (FftComplex(data, data, size, fftscale) != EDSP_BAD_ARG))
		return 1;
    }
    sprintf(rcode, "Testing argument checking of real FFT's\n");
    for (size = -1; size < MIN_RFFT_SIZE; size++)
    {
	if (debug > 255)
	    printf("testing real forward size %ld\n", size);
	if ((FftInReal(data, size, fftscale) != EDSP_BAD_ARG) ||
	    (FftReal(data, data, size, fftscale) != EDSP_BAD_ARG))
		return 1;
    }

    /*
     * Check large transforms
     */
    for (scale_idx = 0; scale_idx < NO_SCALES; scale_idx++)
    {

	fftscale = scales[scale_idx];
	for (size = MIN_CFFT_SIZE; size <= MAX_FFT_SAMP; size *= 2)
	{
	sprintf(rcode, "Testing rejection of size %ld\n", size);
	    /*
	     * Check all non powers of two
	     */
	    for (j = size + 1; j < 2 * size; j++)
	    {
		sprintf(rcode, "Testing rejection of size %ld\n", size);
		if (FftInComplex(data, j, fftscale) != EDSP_BAD_ARG)
		    return 1;
		if (FftComplex(data, data, j, fftscale) != EDSP_BAD_ARG)
		    return 1;
		if (FftInReal(data, j, fftscale) != EDSP_BAD_ARG)
		    return 1;
		if (FftReal(data, data, j, fftscale) != EDSP_BAD_ARG)
		    return 1;
	    }

	    /*
	     * Check that powers of two are processed correctly
	     */
	    printf("Testing FFTs of size %ld\n", size);
	    ipscale = SCALE_BASE * GetCHalving(size, fftscale) / size;
	    result += TestOneFft(rcode, size, fftscale, ipscale);
	    printf("Error count now %d\n", result);
	    if (result)
		return result;
	}
    }

    sprintf(rcode, "TestFft had %d errors\n", result);
    return result;
}

static double
sqr(double d)
{
    return d * d;
}

/*
** FUNCTION:    SingleCompare
**
** AUTHOR:      Ensigma Ltd.
**
** DESCRIPTION: This function compares a single element of the input
**              array with a given value, and all the other elements with
**              zero, using a loose tolerance.
**
** PARAMETERS:  op1 - array to be compared
**              n   - number of elements in the array
**              m   - index for non-zero element
**              singval - required value of non-zero elemnt
**
** RETURNS:     0 if array is zero and given element has the specified value,
**              1 if either is not the case, by more than the tolerance.
**
*/
static int
SingleCompare(short *op1, long n, long m, double singval)
{
#define S_TOLERANCE       0.0022

    int i;
    double rmserror = 0;

    if (debug > 255)
    {
	printf("OP should be %g in bin %ld\n", singval, m);
	for (i = 0; i < n; i++)
	    printf("%d  %hd\n", i, op1[i]);
    }

    for (i = 0; i < m; i++)
	rmserror += sqr(*op1++);
    rmserror += sqr(*op1++ - singval);
    for (i++; i < n; i++)
	rmserror += sqr(*op1++);
    rmserror = sqrt(rmserror / n);

    if (debug)
	printf("Error is %g\n", rmserror / singval);
    return (rmserror / singval >= S_TOLERANCE);
#undef S_TOLERANCE
}
