/*
** FILE NAME:   tgwn.c
**
** PROJECT ID:  94-0040
**
** TITLE:       File that tests the DSPLIB GenGWnoise() function
**
** AUTHOR:      Ensigma Ltd.
**
** DESCRIPTION: The function Tgwn() checks that GenGWnoise()
**              a) Does not corrupt the stack
**              b) Correctly flags incorrect arguments
**              c) Returns sufficiently 'White' Gaussian noise
**              NB: Due to differences in random number generators it
**              may be possible that the test will slightly exceed the
**              declared limits.
**
** PARAMETERS:
**              char *rcode     : pointer to return pass/fail information
**
** MODIFICATION HISTORY:
**
**              1.00    15-Sep-95       Initial Revision
**
*/
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ensigdsp.h>
#include "testhead.h"

#define NARGS   3

#define MAXG 4.5                /* approx. saturating level for N(0,1) random
				 * variable */
#define N_SAMP 256              /* number of samples generated in a frame */
#define N_FRAMES 100            /* number of frames generated */
#define CHISQUARED_LIMIT 67.5   /* 99% Chi-square value for 43 deg. of freedom */
#define SPECTRUM_DEVIATION 3.2  /* limit (dB) on deviation over noise spectrum */
#define N_BINS 44               /* number of histogram bins for analysis of
				 * distribution */
#define BIG     1e20

#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr

/*
** FUNCTION:    Four1
**
** DESCRIPTION: FFT
**
*/
void 
Four1(float data[], int nn, int isign)
{
    int n, mmax, m, j, istep, i;
    double wtemp, wr, wpr, wpi, wi, theta;
    float tempr, tempi;

    isign = -isign;             /* sign change */

    n = nn << 1;
    j = 1;
    for (i = 1; i < n; i += 2)
    {
	if (j > i)
	{
	    SWAP(data[j], data[i]);
	    SWAP(data[j + 1], data[i + 1]);
	}
	m = n >> 1;
	while (m >= 2 && j > m)
	{
	    j -= m;
	    m >>= 1;
	}
	j += m;
    }
    mmax = 2;
    while (n > mmax)
    {
	istep = 2 * mmax;
	theta = 6.28318530717959 / (isign * mmax);
	wtemp = sin(0.5 * theta);
	wpr = -2.0 * wtemp * wtemp;
	wpi = sin(theta);
	wr = 1.0;
	wi = 0.0;
	for (m = 1; m < mmax; m += 2)
	{
	    for (i = m; i <= n; i += istep)
	    {
		j = i + mmax;
		tempr = wr * data[j] - wi * data[j + 1];
		tempi = wr * data[j + 1] + wi * data[j];
		data[j] = data[i] - tempr;
		data[j + 1] = data[i + 1] - tempi;
		data[i] += tempr;
		data[i + 1] += tempi;
	    }
	    wr = (wtemp = wr) * wpr - wi * wpi + wr;
	    wi = wi * wpr + wtemp * wpi + wi;
	}
	mmax = istep;
    }
}

/*
** FUNCTION:    TestProperties
**
** DESCRIPTION: The function runs the white noise generator repeatedly,
**              and measures the distribution and spectrum of the samples.
**              The deviation of the spectrum from Gaussian and the
**              colouration of the spectrum are measured and checked against
**              performance targets.
**
** PARAMETERS:
**              float   var     : average variance of noise samples
**              char    *rcode  : error message string
**
** RETURNS:     0               : OK
**              > 0             : Failed
*/
int 
TestProperties(float var,char *rcode)
{
    int i, j, k;
    short out[N_SAMP];
    float maxval;
    float minval;
    float maxdiff;
    float fftip[2 * N_SAMP + 1];

    long int no[N_BINS];        /* Histogram of sample sizes */
    float pg[N_BINS];           /* Expected number of samples in each bin */
    float chisquared;           /* parameter for goodness of fit test */
    float spect[N_SAMP / 2];    /* mean spectrum */
    int fail = 0;

    static float ival[N_BINS - 1] =     /* Boundaries defining histogram bins
					 * for N(0,1) */
    {-3.0, -2.8, -2.6, -2.4, -2.2, -2.0, -1.8,
     -1.6, -1.4, -1.2, -1.1, -1.0, -0.9, -0.8, -0.7, -0.6, -0.5,
     -0.4, -0.3, -0.2, -0.1, 0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
     1.1, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0};
    static float cg[N_BINS / 2] =       /* Gaussian cumulative values at each
					 * boundary */
    {.0013, .0026, .0047, .0082, .0139,
     .0228, .0359, .0548, .0808, .1151, .1357,
     .1587, .1841, .2119, .2420, .2743, .3085, .3446, .3821, .4207, .4602,
     .5000};

/*
**  determine expected number of hits per bin by subtracting adjacent
**  points of Gaussian cumulative array and scaling
*/
    pg[0] = cg[0] * N_FRAMES * N_SAMP;
    pg[N_BINS - 1] = pg[0];
    for (i = 1; i < N_BINS / 2; i++)
    {
	pg[i] = (cg[i] - cg[i - 1]) * N_FRAMES * N_SAMP;
	pg[N_BINS - 1 - i] = pg[i];
    }
/*
**  initialise histogram bins and mean spectrum
*/
    for (i = 0; i < N_BINS; i++)
	no[i] = 0;
    for (i = 0; i < N_SAMP / 2; i++)
	spect[i] = 0;
/*
**  generate noise, classify each sample, and compute spectrum
*/

    for (i = 0; i < N_FRAMES; i++)
    {
	GenGWnoise(out, N_SAMP, var);
	for (j = 0; j < N_SAMP; j++)
	{
	    for (k = 0; k < N_BINS - 1; k++)
		if (MAXG * out[j] / 32768 < ival[k])
		{
		    break;
		}
	    no[k]++;
	}
/*
** Update the spectral parameters
*/
	for (j = 0; j < N_SAMP; j++)
	{
	    fftip[2 * j + 1] = out[j];  /* fill every 2nd index from 2nd
					 * element onwards */
	    fftip[2 * j + 2] = 0;
	}

	Four1(fftip, N_SAMP, 1);
	for (j = 1; j < N_SAMP / 2; j++)        /* don't use DC or negative
						 * freqs */
	    spect[j] += sqrt(fftip[2 * j + 1] * fftip[2 * j + 1] + fftip[2 * j + 2] * fftip[2 * j + 2]);

    }
/*
** Process the FFT data
*/
    maxval = -BIG;
    minval = BIG;
    for (i = 1; i < N_SAMP / 2; i++)
    {
	if (spect[i] > maxval)
	    maxval = spect[i];
	if (spect[i] < minval)
	    minval = spect[i];
    }
    maxdiff = 20 * log10(maxval / minval);      /* spectral deviation in dBs */

/*
** Now do the test for goodness of fit
*/
    chisquared = 0;
    for (k = 0; k < N_BINS; k++)
	chisquared += (no[k] - pg[k]) * (no[k] - pg[k]) / pg[k];

/*
** See if the figures are acceptable
*/
    if (chisquared > CHISQUARED_LIMIT)
	fail++;                 /* distribution is most likely not pure
				 * Gaussian */
    if (maxdiff > SPECTRUM_DEVIATION)
	fail++;                 /* spectrum is not white */

    if(fail)
	sprintf(rcode,"\nChiSq=%2.2f (Lim=%2.2f), Spectrum Dev.=%2.2fdB (Lim=%2.2f)",chisquared,CHISQUARED_LIMIT,maxdiff,SPECTRUM_DEVIATION);
    return (fail);
}

void 
Tgwn(char *rcode)
{
    int size, errc,x;
    float var;
    short out[N_SAMP];
    static int stkchk[NARGS];

/* test working of stack */
    size = N_SAMP;
    var = 32768;
    stkchk[0] = (int)out;
    stkchk[1] = size;
    stkchk[2] = (int)var;

    x = 0;      /*** 0 for no stack test, 1 for stack test ***/

    if( x == 1)
    {
    /* also test for OK return argument */
    if (GenGWnoise(out, size, var) != EDSP_OK)
	sprintf(rcode, "EDSP_OK not returned\n");
    if (stkchk[0] != (int)out ||
	stkchk[1] != size ||
	stkchk[2] != (int)var)
    {
	sprintf(rcode, "Function arguments altered - aborting");
	exit(1);
    }
    }
/* test return args */
    for (size = -2; size < 1; size++)
	if (GenGWnoise(out, size, 1) != EDSP_BAD_ARG)
	    sprintf(rcode, "Bad Return Argument\n");
    for (var = -1.0; var <= 0.0; var += 1.0)
	if (GenGWnoise(out, N_SAMP, var) != EDSP_BAD_ARG)
	    sprintf(rcode, "Bad Return Argument\n");

/* N(0,1)/MAXG random variables lie approximately in range -1..1, hence
   generate random variables within this range that use the full 16 bits */
    srand(4);
    var = 32768 / MAXG * 32768 / MAXG;
    errc = TestProperties(var,rcode);
    if (strlen(rcode) == 0)
	sprintf(rcode, "GenGWnoise Passed\n");
    else
	strcat(rcode, "GenGWnoise Failed\n");
}
