/*
** FILE NAME:   tstwin.c
**
** PROJECT ID:  94-0040
**
** TITLE:       File that tests the DSPLIB window functions
**
** AUTHOR:      Ensigma Ltd.
**
** DESCRIPTION: The function Tstwin() checks that each of the functions
**              GenBlackman(), GenHanning(), GenHamming() and GenTriangle()
**              a) Correctly flags incorrect arguments
**              b) Tests the floating point library against precalculated data
**              c) Tests the quantisation noise is within tolerance
**              d) Passes a simple test vector for bit-exactness (within tol.)
**
** PARAMETERS:
**              char *rcode     : pointer to return pass/fail information
**
** MODIFICATION HISTORY:
**
**              1.00    15-Sep-95       Initial Revision
**
*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <ensigdsp.h>
#include "testhead.h"

#ifndef DBL_EPSILON             /* account for alternative DBL_EPSILON */
#define DBL_EPSILON     DBL_POS_EPS
#endif

#define FAIL    0
#define PASS    1
#define SC      32767
#define MAXN    2048            /* maximum size of window */
#define NTEST   32              /* length of precalc'd test */
#define NRTEST  NTEST-1         /* length of rounding test */
#define PREC    100.0*SC*DBL_EPSILON    /* suitable machine accuracy */
#define FIXERR  10*log10(1.0/12)+0.75   /* (dB) variance of error = delta^2/12,
					 * allow a */
 /* 0.75 dB tolerance due to noise correlation */
#define FLTERR  -65             /* (dB) suitable threshold for floating point
				 * calcs */
#define TWOPI   6.28318530717959
	float   epsilon;
	float   dave;
	double   error;
	float   poseps;
/*
** FUNCTION:    validate1
**
** DESCRIPTION: Check tolerance of floating point function
**
*/
int 
validate1(double err)
{
    if (err > FLTERR)
	return (FAIL);
    else
	return (PASS);
}

/*
** FUNCTION:    validate1
**
** DESCRIPTION: Check tolerance of fixed point error
**
*/
int 
validate2(double err)
{
    if (err > FIXERR)
	return (FAIL);
    else
	return (PASS);
}

/*
** FUNCTION:    Blackman, Hamming, Hanning, Triangle
**
** DESCRIPTION: Window function as defined
**
*/
double 
Blackman(int n, int len)
{
    return (SC * (0.42 - 0.5 * cos(TWOPI * n / len) + 0.08 * cos(2 * TWOPI * n / len)));
}

double 
Hamming(int n, int len)
{
    return (SC * (0.54 - 0.46 * cos(TWOPI * n / len)));
}

double 
Hanning(int n, int len)
{
    return (SC * 0.5 * (1.0 - cos(TWOPI * n / len)));
}

double 
Triangle(int n, int len)
{
    return (SC * (1.0 - fabs((2.0 * n - len + 1) / (len + 1.0))));
}

/*
** FUNCTION:    roundtest
**
** DESCRIPTION: Determine the rounding (quantisation) noise of the fixed point
**              window function in comparison with a floating point version.
**
*/
double 
roundtest(int len, int (*fixpnt) (short *out, long n), double (*flpnt) (int n, int len))
{
    short output[MAXN];
    int n;
    double temp, err = 0;

    fixpnt(output, len);
    for (n = 0; n < len; n++)
    {
	temp = flpnt(n, len);
	err += (output[n] - temp) * (output[n] - temp);
    }
    return (10 * log10(err / len));
}

void 
Tstwin(char *rcode)
{
    short output[MAXN];
    long len, n;
    double temp, err1, err2;
#include "tstwin.txt"           /* load test data for N=32 */

	poseps  =       DBL_POS_EPS;
	epsilon =       DBL_EPSILON;
	dave    =       PREC;
	error   =       (0.5 + PREC);
/* check window length */
    for (len = -2; len <= 1; len++)
    {
	if ((GenBlackman(output, len)) != EDSP_BAD_ARG)
	    sprintf(rcode, "Bad argument in GenBlackman\n");
	if ((GenHanning(output, len)) != EDSP_BAD_ARG)
	    sprintf(rcode, "Bad argument in GenBlackman\n");
	if ((GenHamming(output, len)) != EDSP_BAD_ARG)
	    sprintf(rcode, "Bad argument in GenBlackman\n");
	if ((GenTriangle(output, len)) != EDSP_BAD_ARG)
	    sprintf(rcode, "Bad argument in GenBlackman\n");
    }

/* Compare floating point window generator functions against precalculated data
   to verify this file's floating point library.  Use this to test library
   function */

    /* Blackman */
    len = NTEST;
    err1 = 0;
    err2 = 0;
    GenBlackman(output, len);
    for (n = 0; n < len; n++)
    {
	temp = Blackman(n, len);
	err1 += (wbla[n] - temp) * (wbla[n] - temp);
    }
    err1 = 10 * log10(err1 / len);
    if (!validate1(err1))
	sprintf(rcode, "Blackman failed float test\n");
    err2 = roundtest(NRTEST, &GenBlackman, &Blackman);
    if (!validate2(err2))
	sprintf(rcode, "Blackman failed rounding test\n");

    /* Hamming */
    len = NTEST;
    err1 = 0;
    err2 = 0;
    GenHamming(output, len);
    for (n = 0; n < len; n++)
    {
	temp = Hamming(n, len);
	err1 += (wham[n] - temp) * (wham[n] - temp);
    }
    err1 = 10 * log10(err1 / len);
    if (!validate1(err1))
	sprintf(rcode, "Hamming failed float test\n");
    err2 = roundtest(NRTEST, &GenHamming, &Hamming);
    if (!validate2(err2))
	sprintf(rcode, "Hamming failed rounding test\n");

    /* Hanning */
    len = NTEST;
    err1 = 0;
    err2 = 0;
    GenHanning(output, len);
    for (n = 0; n < len; n++)
    {
	temp = Hanning(n, len);
	err1 += (whan[n] - temp) * (whan[n] - temp);
    }
    err1 = 10 * log10(err1 / len);
    if (!validate1(err1))
	sprintf(rcode, "Hanning failed float test\n");
    err2 = roundtest(NRTEST, &GenHanning, &Hanning);
    if (!validate2(err2))
	sprintf(rcode, "Hanning failed rounding test\n");

    /* Triangle */
    len = NTEST;
    err1 = 0;
    err2 = 0;
    GenTriangle(output, len);
    for (n = 0; n < len; n++)
    {
	temp = Triangle(n, len);
	err1 += (wtri[n] - temp) * (wtri[n] - temp);
    }
    err1 = 10 * log10(err1 / len);
    if (!validate1(err1))
	sprintf(rcode, "Triangle failed float test\n");
    err2 = roundtest(NRTEST, &GenTriangle, &Triangle);
    if (!validate2(err2))
	sprintf(rcode, "Triangle failed rounding test\n");

/* check small window lengths on a bit by bit basis
   NB a tolerance of 0.5 bit is permitted (within machine accuracy) */

    for (len = 28; len < 30; len++)
    
	{
	GenHanning(output, len);
	for (n = 7; n < len; n++)
		error = (fabs(Hanning(n, len) - output[n]));
	    if (fabs(Hanning(n, len) - output[n]) > (0.5 + PREC))
		printf(" HN Failed ");
	
	GenBlackman(output, len);
	for (n = 0; n < len; n++)
	    if (fabs(Blackman(n, len) - output[n]) > (0.5 + PREC))
		sprintf(rcode, "BL Failed\n");
	
	GenHamming(output, len);
	for (n = 0; n < len; n++)
	    if (fabs(Hamming(n, len) - output[n]) > (0.5 + PREC))
		sprintf(rcode, "HM Failed\n");
	
	GenTriangle(output, len);
	for (n = 0; n < len; n++)
	    if (fabs(Triangle(n, len) - output[n]) > (0.5 + PREC))
		sprintf(rcode, "TR Failed\n");
    }

    if (strlen(rcode) == 0)
	sprintf(rcode, "Window Passed\n");
    else
	strcat(rcode, "Window Failed\n");
}
